# mypy: ignore-errors
import gliner2
import pydantic
import pytest
from flaky import flaky

from sieves import Doc, Pipeline, tasks
from sieves.model_wrappers import ModelType, ModelSettings, dspy_, langchain_, outlines_
from sieves.serialization import Config
from sieves.tasks import PredictiveTask, InformationExtraction
from sieves.tasks.predictive import information_extraction
from sieves.tasks.predictive.schemas.information_extraction import ResultSingle, ResultMulti


class Person(pydantic.BaseModel, frozen=True):
    name: str
    age: pydantic.PositiveInt
    score: pydantic.NonNegativeFloat | None = None

class PersonNotFrozen(pydantic.BaseModel):
    name: str
    age: pydantic.PositiveInt
    score: pydantic.NonNegativeFloat | None = None

@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize(
    "batch_runtime",
    InformationExtraction.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
@pytest.mark.parametrize("mode", ["multi", "single"])
def test_run(information_extraction_docs, batch_runtime, fewshot, mode) -> None:
    if mode == "multi":
        fewshot_examples = [
            information_extraction.FewshotExampleMulti(
                text="Ada Lovelace lived to 47 years old. Zeno of Citium died with 72 years.",
                entities=[Person(name="Ada Lovelace", age=47, score=1.), Person(name="Zeno of Citium", age=72)],
            ),
            information_extraction.FewshotExampleMulti(
                text="Alan Watts passed away at the age of 58 years. Alan Watts was 58 years old at the time of his death.",
                entities=[Person(name="Alan Watts", age=58)],
            ),
    ]

    else:
        fewshot_examples = [
            information_extraction.FewshotExampleSingle(
                text="Ada Lovelace lived to 47 years old.",
                entity=Person(name="Ada Lovelace", age=47, score=1.),
            ),
            information_extraction.FewshotExampleSingle(
                text="Alan Watts passed away at the age of 58 years.",
                entity=Person(name="Alan Watts", age=58),
            ),
        ]

    entity_type = Person
    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = tasks.predictive.InformationExtraction(
        entity_type=entity_type,
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        mode=mode,
        **fewshot_args
    )

    pipe = Pipeline(task)
    docs = list(pipe(information_extraction_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "InformationExtraction" in doc.results

        # Verify unified result types.
        res = doc.results["InformationExtraction"]
        if mode == "multi":
            assert isinstance(res, information_extraction.ResultMulti)
            assert len(res.entities)
        else:
            assert isinstance(res, information_extraction.ResultSingle)
            assert res.entity is not None

        assert "InformationExtraction" in doc.meta
        assert "raw" in doc.meta["InformationExtraction"]
        assert "usage" in doc.meta["InformationExtraction"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['InformationExtraction']}")
        print(f"Raw output: {doc.meta['InformationExtraction']['raw']}")
        print(f"Usage: {doc.meta['InformationExtraction']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

        if mode == "single":
            assert (
                doc.results["InformationExtraction"].entity is None or
                isinstance(doc.results["InformationExtraction"].entity, pydantic.BaseModel)
            )
        else:
            assert isinstance(doc.results["InformationExtraction"].entities, list)

    with pytest.raises(NotImplementedError):
        pipe["InformationExtraction"].distill(None, None, None, None, None, None, None, None)

    if fewshot_examples:
        _to_hf_dataset(task, docs, mode)


def _to_hf_dataset(task: InformationExtraction, docs: list[Doc], mode: str) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: InformationExtraction task instance.
    :param docs: List of documents to convert.
    :param mode: Extraction mode.
    """
    assert isinstance(task, PredictiveTask)
    dataset = task.to_hf_dataset(docs)
    target_field = "entities" if mode == "multi" else "entity"
    assert all([key in dataset.features for key in ("text", target_field)])
    assert len(dataset) == 2
    records = list(dataset)
    assert records[0]["text"] == "Mahatma Ghandi lived to 79 years old. Bugs Bunny is at least 85 years old."
    assert records[1]["text"] == "Marie Curie passed away at the age of 67 years. Marie Curie was 67 years old."
    for record in records:
        res = record[target_field]
        if mode == "multi":
            assert isinstance(res, dict)
            assert isinstance(res["age"], list)
            assert isinstance(res["name"], list)
        else:
            assert res is None or isinstance(res, dict)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])

@pytest.mark.parametrize(
    "batch_runtime",
    [ModelType.outlines],
    indirect=["batch_runtime"],
)
def test_frozen_check(batch_runtime) -> None:
    # This should work (frozen)
    tasks.predictive.InformationExtraction(
        entity_type=Person,
        model=batch_runtime.model,
    )

    # This should raise ValueError (not frozen)
    with pytest.raises(ValueError, match="isn't frozen"):
        tasks.predictive.InformationExtraction(
            entity_type=PersonNotFrozen,
            model=batch_runtime.model,
        )


@pytest.mark.parametrize("batch_runtime", [ModelType.outlines], indirect=["batch_runtime"])
@pytest.mark.parametrize("mode", ["multi", "single"])
def test_serialization(information_extraction_docs, batch_runtime, mode) -> None:
    pipe = Pipeline(
        tasks.predictive.InformationExtraction(
            entity_type=Person,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
            mode=mode,
        )
    )

    config = pipe.serialize()
    assert config.model_dump() == {
        'cls_name': 'sieves.pipeline.core.Pipeline',
        'tasks': {
            'is_placeholder': False,
            'value': [{
                'cls_name': 'sieves.tasks.predictive.information_extraction.core.InformationExtraction',
                'entity_type': {'is_placeholder': True, 'value': 'pydantic._internal._model_construction.ModelMetaclass'},
                'fewshot_examples': {'is_placeholder': False, 'value': ()},
                'batch_size': {'is_placeholder': False, "value": -1},
                'mode': {'is_placeholder': False, 'value': mode},
                'model_settings': {
                    'is_placeholder': False,
                    'value': {
                        'config_kwargs': None,
                        'inference_kwargs': None,
                        'init_kwargs': None,
                        'strict': True,
                        'inference_mode': None
                    }
                },
                'include_meta': {'is_placeholder': False, 'value': True},
                'model': {'is_placeholder': True, 'value': 'outlines.models.openai.OpenAI'},
                'prompt_instructions': {'is_placeholder': False, 'value': None},
                'task_id': {'is_placeholder': False, 'value': 'InformationExtraction'},
                'condition': {'is_placeholder': False, 'value': None},
                'version': Config.get_version()
            }]
        },
        'use_cache': {'is_placeholder': False, 'value': True},
        'version': Config.get_version()
    }

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model, "entity_type": Person, "mode": mode}])


@pytest.mark.parametrize(
    "batch_runtime",
    InformationExtraction.supports(),
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = tasks.predictive.InformationExtraction(
        entity_type=Person,
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize("batch_runtime", [ModelType.outlines], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for information extraction without running pipeline."""
    task = tasks.predictive.InformationExtraction(entity_type=Person, model=batch_runtime.model, task_id="ie", mode="single")

    # 1. Full overlap (Single)
    doc_full = Doc(text="Ada is 47.")
    res_full = ResultSingle(entity=Person(name="Ada", age=47))
    doc_full.results["ie"] = res_full
    doc_full.gold["ie"] = res_full
    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. No overlap (Single)
    doc_none = Doc(text="Ada is 47.")
    res_none_pred = ResultSingle(entity=Person(name="Alan", age=58))
    doc_none.results["ie"] = res_none_pred
    doc_none.gold["ie"] = res_full
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0

    # 3. Multi-label partial overlap
    task_multi = tasks.predictive.InformationExtraction(
        entity_type=Person, model=batch_runtime.model, task_id="ie_multi", mode="multi"
    )
    doc_partial = Doc(text="Ada is 47 and Alan is 58.")
    # Pred has both
    res_multi_pred = ResultMulti(entities=[Person(name="Ada", age=47), Person(name="Alan", age=58)])
    # Gold has only Ada
    res_multi_gold = ResultMulti(entities=[Person(name="Ada", age=47)])
    doc_partial.results["ie_multi"] = res_multi_pred
    doc_partial.gold["ie_multi"] = res_multi_gold
    report_partial = task_multi.evaluate([doc_partial])
    # TP=1, FP=1, FN=0 -> Precision=0.5, Recall=1.0 -> F1=0.666...
    assert 0.6 < report_partial.metrics[task_multi.metric] < 0.7



# mypy: ignore-errors
import pytest
from flaky import flaky

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType, ModelSettings
from sieves.serialization import Config
from sieves.tasks.predictive import ner
from sieves.tasks.predictive.schemas.ner import EntityWithContext, TaskResult, Entity


@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize(
    "batch_runtime",
    ner.NER.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(ner_docs, batch_runtime, fewshot) -> None:
    fewshot_examples = [
        ner.FewshotExample(
            text="John studied data science in Barcelona and lives with Jaume",
            entities=[
                EntityWithContext(text="John", context="John studied data", entity_type="PERSON", score=1.0),
                EntityWithContext(text="Barcelona", context="science in Barcelona", entity_type="LOCATION", score=1.0),
                EntityWithContext(text="Jaume", context="lives with Jaume", entity_type="PERSON", score=1.0),
            ],
        ),
        ner.FewshotExample(
            text="Maria studied computer engineering in Madrid and works with Carlos",
            entities=[
                EntityWithContext(text="Maria", context="Maria studied computer", entity_type="PERSON", score=1.0),
                EntityWithContext(text="Madrid", context="engineering in Madrid and works", entity_type="LOCATION", score=1.0),
                EntityWithContext(text="Carlos", context="works with Carlos", entity_type="PERSON", score=1.0),
            ],
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = ner.NER(
        entities=["PERSON", "LOCATION", "COMPANY"],
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        **fewshot_args
    )
    pipe = Pipeline(task)
    docs = list(pipe(ner_docs))

    assert len(docs) == 2
    for doc in docs:
        assert "NER" in doc.results

        # Verify unified result types.
        assert isinstance(doc.results["NER"], ner.Result)

        assert "NER" in doc.meta
        assert "raw" in doc.meta["NER"]
        assert "usage" in doc.meta["NER"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['NER']}")
        print(f"Raw output: {doc.meta['NER']['raw']}")
        print(f"Usage: {doc.meta['NER']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    with pytest.raises(NotImplementedError):
        pipe["NER"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(ner_docs, batch_runtime) -> None:
    pipe = Pipeline(
        ner.NER(
            entities=["PERSON", "LOCATION", "COMPANY"],
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    )

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
 'tasks': {'is_placeholder': False,
           'value': [{'cls_name': 'sieves.tasks.predictive.ner.core.NER',
                      'entities': {'is_placeholder': False,
                                   'value': ['PERSON', 'LOCATION', 'COMPANY']},
                      'fewshot_examples': {'is_placeholder': False,
                                           'value': ()},
                      'batch_size': {'is_placeholder': False, "value": -1},
                      'model_settings': {'is_placeholder': False,
                                              'value': {
                                                        'config_kwargs': None,
                                                        'inference_kwargs': None,
                                                        'init_kwargs': None,
                                                        'strict': True,
                                                        'inference_mode': None,}},
                      'include_meta': {'is_placeholder': False, 'value': True},
                      'model': {'is_placeholder': True,
                                'value': 'dspy.clients.lm.LM'},
                      'prompt_instructions': {'is_placeholder': False,
                                          'value': None},
                      'task_id': {'is_placeholder': False, 'value': 'NER'},
                      'condition': {'is_placeholder': False, 'value': None},
                      'version': Config.get_version()}]},
 'use_cache': {'is_placeholder': False, 'value': True},
 'version': Config.get_version()}
    Pipeline.deserialize(
        config=config,
        tasks_kwargs=[{"model": batch_runtime.model}],
    )


def _to_hf_dataset(task: ner.NER, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: NER task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "entities")])
    assert len(dataset) == 2
    dataset_records = list(dataset)
    for rec in dataset_records:
        assert isinstance(rec["entities"], dict)
        assert (
            len(rec["entities"]["entity_type"])
            == len(rec["entities"]["start"])
            == len(rec["entities"]["end"])
            == len(rec["entities"]["text"])
        )
        assert isinstance(rec["text"], str)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize(
    "batch_runtime",
    [ModelType.dspy, ModelType.langchain, ModelType.outlines, ModelType.gliner],
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = ner.NER(
        entities=["PERSON", "LOCATION", "COMPANY"],
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize(
    "batch_runtime",
    ner.NER.supports(),
    indirect=["batch_runtime"],
)
def test_run_with_dict_entities(ner_docs, batch_runtime) -> None:
    """Test NER with dict format entities (labels with descriptions)."""
    entities_with_descriptions = {
        "PERSON": "Names of people",
        "LOCATION": "Geographic locations",
        "COMPANY": "Names of companies and organizations"
    }

    pipe = Pipeline(
        ner.NER(
            entities=entities_with_descriptions,
            model=batch_runtime.model,
            model_settings=batch_runtime.model_settings,
            batch_size=batch_runtime.batch_size,
        )
    )
    docs = list(pipe(ner_docs))

    assert len(docs) == 2
    for doc in docs:
        assert "NER" in doc.results


@pytest.mark.parametrize("batch_runtime", [ModelType.gliner], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for NER without running pipeline."""
    task = ner.NER(entities=["PERSON", "LOCATION"], model=batch_runtime.model, task_id="ner")

    # 1. Full overlap
    doc_full = Doc(text="John lives in Berlin.")
    res_full = TaskResult(
        text=doc_full.text,
        entities=[
            Entity(text="John", start=0, end=4, entity_type="PERSON"),
            Entity(text="Berlin", start=14, end=20, entity_type="LOCATION")
        ]
    )
    doc_full.results["ner"] = res_full
    doc_full.gold["ner"] = res_full
    report_full = task.evaluate([doc_full])
    assert report_full.metrics[task.metric] == 1.0

    # 2. Partial overlap
    doc_partial = Doc(text="John lives in Berlin.")
    # Pred has John and Berlin
    res_pred = TaskResult(
        text=doc_partial.text,
        entities=[
            Entity(text="John", start=0, end=4, entity_type="PERSON"),
            Entity(text="Berlin", start=14, end=20, entity_type="LOCATION")
        ]
    )
    # Gold has John and Paris (Berlin is missing, Paris is extra in gold)
    res_gold = TaskResult(
        text=doc_partial.text,
        entities=[
            Entity(text="John", start=0, end=4, entity_type="PERSON"),
            Entity(text="Paris", start=14, end=19, entity_type="LOCATION")
        ]
    )
    doc_partial.results["ner"] = res_pred
    doc_partial.gold["ner"] = res_gold
    report_partial = task.evaluate([doc_partial])
    # Precision: 1 TP (John) / 2 Pred = 0.5
    # Recall: 1 TP (John) / 2 Gold = 0.5
    # F1: 0.5. (Note: _evaluate_dspy_example uses the logic in NER subclass which computes F1)
    assert report_partial.metrics[task.metric] == 0.5

    # 3. No overlap
    doc_none = Doc(text="John lives in Berlin.")
    res_none_pred = TaskResult(
        text=doc_none.text, entities=[Entity(text="John", start=0, end=4, entity_type="PERSON")]
    )
    res_none_gold = TaskResult(
        text=doc_none.text, entities=[Entity(text="Berlin", start=14, end=20, entity_type="LOCATION")]
    )
    doc_none.results["ner"] = res_none_pred
    doc_none.gold["ner"] = res_none_gold
    report_none = task.evaluate([doc_none])
    assert report_none.metrics[task.metric] == 0.0



"""Ingestion tasks package exports.

Re-export tasks through an import wrapper that gracefully handles optional
dependencies (docling, marker).
"""

from sieves.tasks.preprocessing.ingestion.ingestion_import import Docling, Marker, docling_, marker_  # isort: skip
from sieves.tasks.preprocessing.ingestion.core import Ingestion

__all__ = [
    "Ingestion",
    "Docling",
    "Marker",
    "docling_",
    "marker_",
]



import marimo

__generated_with = "0.18.4"
app = marimo.App(width="medium")


@app.cell
def _():
    import marimo as mo
    return (mo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Case Study: Structured Crisis Tweet Analysis

    This notebook demonstrates how to use `sieves` to build a structured information extraction pipeline. We'll use a **toy example** involving crisis-related tweets to show how `sieves` can help sift through unstructured text to identify relevant events and extract key entities.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    The goal is to build a multi-stage filter (a sieve, if you will) that:
    1. **Classifies** if a tweet is relevant.
    2. **Extracts** the crisis type and location **only if** it's relevant.

    This conditional orchestration allows for efficient processing and reduced noise.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Downloading the Data


    We'll first download a [dataset](https://crisisnlp.qcri.org/data/lrec2016/labeled_cf/CrisisNLP_labeled_data_crowdflower_v2.zip) from https://crisisnlp.qcri.org/.
    """)
    return


@app.cell
def _():
    import urllib.request
    import zipfile
    from pathlib import Path
    import shutil

    url = "https://crisisnlp.qcri.org/data/lrec2016/labeled_cf/CrisisNLP_labeled_data_crowdflower_v2.zip"
    zip_path = Path("CrisisNLP_dataset.zip")
    data_dir = Path("CrisisNLP")

    if not zip_path.exists():
        urllib.request.urlretrieve(url, str(zip_path))

    with zipfile.ZipFile(str(zip_path), 'r') as z:
        z.extractall(".")

    extracted = Path("CrisisNLP_labeled_data_crowdflower")
    data_dir.mkdir(exist_ok=True)
    for item in extracted.iterdir():
        target = data_dir / item.name
        if not target.exists():
            item.rename(target)
    shutil.rmtree(extracted)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Next we'll load the data into memory and preprocess it.
    """)
    return


@app.cell
def _():
    import pandas as pd
    import os

    data: pd.DataFrame = pd.DataFrame()

    for file in os.listdir("CrisisNLP"):
        if os.path.isdir(os.path.join("CrisisNLP", file)):
            for file_ in os.listdir(os.path.join("CrisisNLP", file)):

                if file_.endswith(".tsv"):
                    data_ = pd.read_csv(os.path.join("CrisisNLP", file, file_), sep="\t")
                    if len(data_) == 0:
                        continue

                    data_["dataset"] = file.split("_")[1]
                    crisis_type = file.split("_")[2]
                    if crisis_type == 'eq':
                        crisis_type = 'Earthquake'
                    elif crisis_type in ('Odile', 'Pam', 'Typhoon'):
                        crisis_type = 'Hurrican'
                    elif crisis_type in ('ebola', 'cf', 'East'):
                        crisis_type = 'Diseases'
                    elif crisis_type == 'floods':
                        crisis_type = 'Floods'
                    data_["crisis_type"] = crisis_type
                    data = pd.concat([data, data_])
    return data, os


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ### Data Exploration

    We'll load the CrisisNLP dataset, which contains tweets labeled by humans across different disaster types. This "gold" data allows us to evaluate our automated pipeline later.
    """)
    return


@app.cell
def _(data: pd.DataFrame, mo):
    mo.ui.table(data.head(100), label="CrisisNLP Dataset Preview")
    return


@app.cell
def _(data: pd.DataFrame):
    # Quick summary of the dataset
    summary = {
        "Total Tweets": len(data),
        "Crisis Types": data.crisis_type.unique().tolist(),
        "Labels": data.label.unique().tolist()
    }
    summary
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Now we'll convert the data into `sieves` documents, which will allow us to process them. We'll also sample down the data to cut down the time we need for processing the data.
    """)
    return


@app.cell
def _(data: pd.DataFrame):
    from sieves import Doc
    from sieves.tasks.predictive.classification import ResultSingleLabel
    import random

    # We down-sample our dataset to avoid long processing times.
    data_sampled = data.sample(n=100)

    docs = [
        Doc(
            uri=f"tweet_{row['tweet_id']}",
            text=row['tweet_text'],
            meta={
                'tweet_id': row['tweet_id'],
                'dataset': row['dataset'],
            },
            # We store ground truth in .gold per task ID, so we can evaluate our pipeline performance later.
            gold={
                "crisis_label_classifier":  row['label'],
                "crisis_type_classifier": row['crisis_type']
            }
        )
        for idx, row in data_sampled.iterrows()
    ]

    print(f"Created {len(docs)} docs.")
    print(f"\nSample doc:")
    print(f"Tweet: {docs[0].text[:100]}...")
    print(f"Metadata: {docs[0].meta}")
    print(f"Ground truth: {docs[0].gold}")
    return Doc, data_sampled, docs


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Setting up the Pipeline

    A `sieves` pipeline is composed of tasks. We'll use:
    1. **ClassificationTask**: To filter for relevance and identify crisis types.
    2. **InformationExtractionTask**: To extract structured entities (Locations) using Pydantic models.

    ### Conditional Orchestration
    We only want to run the expensive extraction tasks on tweets that are actually related to a crisis. `sieves` allows us to define a `condition` function that acts as a gatekeeper.
    """)
    return


@app.cell
def _(os):
    import dspy

    # We use a lightweight model for this demonstration
    model = dspy.LM(
        "openrouter/google/gemini-2.5-flash-lite-preview-09-2025",
        api_base="https://openrouter.ai/api/v1",
        api_key=os.environ['OPENROUTER_API_KEY'],
    )

    # Batching improves throughput by processing multiple docs in a single prompt
    batch_size = 10
    return batch_size, model


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Let's create a task predicting tweet labels, one of which indicates whether the tweet is being relevant to any crisis:
    """)
    return


@app.cell
def _(batch_size, data_sampled, model):
    from sieves import tasks

    crisis_label_classifier = tasks.Classification(
        task_id="crisis_label_classifier",
        labels=data_sampled.label.unique(),
        mode='single',
        model=model,
        batch_size=batch_size,
    )
    return crisis_label_classifier, tasks


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ### Defining the "Gatekeeper"
    The `related_to_crisis` function checks the output of the first classifier. If the tweet isn't relevant or the confidence is too low, subsequent tasks in the pipeline will be skipped for that document.
    """)
    return


@app.cell
def _(Doc, batch_size, data_sampled, model, tasks):
    def related_to_crisis(doc: Doc) -> bool:
        """Checks if the tweet is relevant enough to proceed with further extraction."""
        result = doc.results.get("crisis_label_classifier")
        if not result:
            return False
        return result.label != 'not_related_or_irrelevant' and result.score >= .6


    crisis_type_classifier = tasks.Classification(
        task_id="crisis_type_classifier",
        labels=data_sampled.crisis_type.unique(),
        mode='single',
        model=model,
        condition=related_to_crisis,
        batch_size=batch_size,
    )
    return crisis_type_classifier, related_to_crisis


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ### Schema-Driven Information Extraction
    By using Pydantic models, we ensure that the LLM output is parsed into a structured object. This makes the data immediately useful for downstream applications.
    """)
    return


@app.cell
def _():
    import pydantic
    from typing import Literal


    class Country(pydantic.BaseModel, frozen=True):
        name: str | None = pydantic.Field(
            description="The name of the country mentioned in the tweet, if any."
        )
    return (Country,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Having the entity types defined, we can proceed to define our tasks. We'll reuse the conditional check from before that allows us to skip irrelevant tweets.
    """)
    return


@app.cell
def _(Country, batch_size, model, related_to_crisis, tasks):
    location_extractor = tasks.InformationExtraction(
        task_id="location_extractor",
        entity_type=Country,
        model=model,
        mode='single',
        batch_size=batch_size,
        condition=related_to_crisis
    )
    return (location_extractor,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Finally we can combine four tasks into a pipeline:
    """)
    return


@app.cell
def _(crisis_label_classifier, crisis_type_classifier, location_extractor):
    pipeline = (
        crisis_label_classifier +
        crisis_type_classifier +
        location_extractor
    )
    return (pipeline,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Evaluating Results

    ### Running the Pipeline

    Everything is in place, so we run our pipeline and collect results:
    """)
    return


@app.cell
def _(docs, pipeline):
    results = list(pipeline(docs))
    return (results,)


@app.cell
def _(results):
    from typing import Any

    # Flatten results for display in a table
    display_results: list[dict[str, Any]] = []
    for doc in results:
        display_results.append({
            "Tweet": doc.text[:100] + "...",
            "Label": doc.results.get('crisis_label_classifier').label if doc.results.get('crisis_label_classifier') else "N/A",
            "Type": doc.results.get('crisis_type_classifier').label if doc.results.get('crisis_type_classifier') else "Skipped",
            "Location": doc.results.get('location_extractor').entity.name if doc.results.get('location_extractor') and doc.results.get('location_extractor').entity else "None/Skipped"
        })
    return (display_results,)


@app.cell
def _(display_results: list[dict[str, Any]], mo):
    mo.ui.table(display_results)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ### Evaluating Pipeline Reliability

    In a real-world scenario, we need to know if we can trust our automated extraction. By comparing our pipeline's "Predicted" results against the "Gold" human labels, we can calculate metrics like Accuracy and F1 Score.

    We do this by running the inbuilt evaluation functionality. Note that we don't have ground truth for the location extraction, so we'll ignore this task in our evaluation.
    """)
    return


@app.cell
def _(pipeline, results):
    from pprint import pprint

    eval_report = pipeline.evaluate(results)
    for task_id in eval_report.reports:
        pprint(eval_report[task_id].summary())
    return (eval_report,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    That seems quite fine for `crisis_type`, and not so great for `crisis_label` - the mediocre performance for labels is probably because there are quite a few of them (15) and the model has difficulties telling them apart.

    Let's inspect the errors - hopefully we can learn from to improve our pipeline (or maybe even find mistakes in the gold data!):
    """)
    return


@app.cell
def _(eval_report, mo):
    from sklearn.metrics import f1_score, accuracy_score, recall_score

    errors: list[dict[str, str | float]] = []

    for tid in ("crisis_label_classifier", "crisis_type_classifier"):
        for failed_pred in eval_report[tid].failures:
            pred = failed_pred.results[tid]
            errors.append({
                "Task": tid,
                "Text": failed_pred.text,
                "Gold": failed_pred.gold[tid],
                "Predicted": pred.label if pred else pred,
            })

    mo.ui.table(errors)
    return


@app.cell(hide_code=True)
def _():
    return


if __name__ == "__main__":
    app.run()



"""PII masking."""

from sieves.tasks.predictive.pii_masking.core import PIIMasking
from sieves.tasks.predictive.schemas.pii_masking import FewshotExample, PIIEntity, Result

__all__ = ["PIIMasking", "FewshotExample", "PIIEntity", "Result"]



"""Schemas for PII masking task."""

from __future__ import annotations

import dspy
import pydantic

from sieves.model_wrappers import dspy_, langchain_, outlines_
from sieves.tasks.predictive.schemas.core import FewshotExample as BaseFewshotExample


class PIIEntity(pydantic.BaseModel, frozen=True):
    """Personally Identifiable Information (PII) entity.

    Attributes:
        entity_type: Type of PII.
        text: Entity text.
        score: Confidence score.
    """

    entity_type: str = pydantic.Field(description="The type of PII identified (e.g., EMAIL, PHONE, SSN).")
    text: str = pydantic.Field(description="The original text of the PII entity.")
    score: float | None = pydantic.Field(
        default=None, description="Provide a confidence score for the PII identification, between 0 and 1."
    )


class FewshotExample(BaseFewshotExample):
    """Example for PII masking few-shot prompting.

    Attributes:
        text: Input text.
        masked_text: Masked version of text.
        pii_entities: List of PII entities.
    """

    masked_text: str
    pii_entities: list[PIIEntity]

    @property
    def target_fields(self) -> tuple[str, ...]:
        """Return target fields.

        :return: Target fields.
        """
        return ("masked_text", "pii_entities")


# --8<-- [start:Result]
class Result(pydantic.BaseModel):
    """Result of a PII masking task. Contains the masked text and the identified PII entities.

    PII entities should be masked with [MASKED].

    Attributes:
        masked_text: Masked version of text.
        pii_entities: List of PII entities.
    """

    masked_text: str = pydantic.Field(description="The original text with PII entities replaced by placeholders.")
    pii_entities: list[PIIEntity] = pydantic.Field(description="List of all PII entities identified in the text.")


# --8<-- [end:Result]


TaskModel = dspy_.Model | langchain_.Model | outlines_.Model
TaskPromptSignature = type[dspy.Signature] | type[pydantic.BaseModel]
TaskResult = Result



# mypy: ignore-errors
import os
import pickle
import tempfile
from pathlib import Path

import chonkie
import dspy
import pytest

from sieves import Pipeline
from sieves.model_wrappers import ModelType
from sieves.serialization import Config
from sieves.tasks import preprocessing
from sieves.tasks.predictive import classification
from sieves.tasks.predictive.classification import ResultMultiLabel
from sieves.tests.conftest import make_model


@pytest.mark.parametrize(
    "batch_runtime",
    [ModelType.dspy],
    indirect=["batch_runtime"],
)
def test_serialization_pipeline(dummy_docs, batch_runtime, tokenizer):
    """Tests serialization and deserialization of pipeline to files and config objects."""
    pipe = Pipeline(
        [
            preprocessing.Chunking(chonkie.TokenChunker(tokenizer)),
            classification.Classification(
                task_id="classifier",
                labels={"science": "Everything about science.", "politics": "Everything about politics."},
                model=batch_runtime.model,
                model_settings=batch_runtime.model_settings,
                batch_size=batch_runtime.batch_size,
            ),
        ]
    )

    # Get config, assert values are correct.
    config = pipe.serialize()
    config_model_dump = config.model_dump()
    version = Config.get_version()
    assert config_model_dump == {
        "cls_name": "sieves.pipeline.core.Pipeline",
        "use_cache": {"is_placeholder": False, "value": True},
        "tasks": {
            "is_placeholder": False,
            "value": [
                {
                    "chunker": {"is_placeholder": True, "value": "chonkie.chunker.token.TokenChunker"},
                    'batch_size': {'is_placeholder': False, 'value': -1},
                    "cls_name": "sieves.tasks.preprocessing.chunking.core.Chunking",
                    "include_meta": {"is_placeholder": False, "value": False},
                    "task_id": {"is_placeholder": False, "value": "Chunking"},
                    'condition': {'is_placeholder': False, 'value': None},
                    "version": version,
                },
                {
                    "cls_name": "sieves.tasks.predictive.classification.core.Classification",
                    'model_settings': {
                        'is_placeholder': False,
                        'value': {
                            'config_kwargs': None,
                            'inference_kwargs': None,
                            'init_kwargs': None,
                            'strict': True,
                            'inference_mode': None,
                        }
                    },
                    'condition': {'is_placeholder': False, 'value': None},
                    "fewshot_examples": {"is_placeholder": False, "value": ()},
                    "include_meta": {"is_placeholder": False, "value": True},
                    'batch_size': {'is_placeholder': False, 'value': -1},
                    "labels": {
                        "is_placeholder": False,
                        "value": {"science": "Everything about science.", "politics": "Everything about politics."},
                    },
                    'model': {'is_placeholder': True, 'value': 'dspy.clients.lm.LM'},
                    "prompt_instructions": {"is_placeholder": False, "value": None},
                    "task_id": {"is_placeholder": False, "value": "classifier"},
                    "version": version,
                },
            ],
        },
        "version": version,
    }

    # Save config to temporary file
    with tempfile.NamedTemporaryFile(suffix=".yml") as tmp_file:
        tmp_path = Path(tmp_file.name)
        config.dump(tmp_path)

        # Load config from file and verify it matches
        loaded_config = Config.load(tmp_path)
        # For some reason empty tuple is stored as list, which is fine for our purposes.
        assert config_model_dump["tasks"]["value"][1]["fewshot_examples"]["value"] == ()
        config_model_dump["tasks"]["value"][1]["fewshot_examples"]["value"] = []
        assert loaded_config.model_dump() == config_model_dump

        # Restore pipeline from config.
        loaded_pipe = Pipeline.load(
            tmp_path,
            (
                {"chunker": chonkie.TokenChunker(tokenizer)},
                {"model": make_model(ModelType.dspy)},
            ),
        )

        # Run restored pipeline.
        docs = list(loaded_pipe(dummy_docs))
        assert len(docs) == 2
        res = docs[0].results["classifier"]
        assert isinstance(res, ResultMultiLabel)
        assert len(res.label_scores)

        # Compare loaded pipe config with original one.
        assert loaded_pipe.serialize().model_dump() == config_model_dump


def test_serialization_docs(dummy_docs):
    """Test serializition of docs by saving to and loading from pickle objects."""
    # Create a temporary file for pickle serialization.
    with tempfile.NamedTemporaryFile(suffix=".pkl") as tmp_file:
        tmp_path = Path(tmp_file.name)

        # Pickle the dummy_docs to file.
        with open(tmp_path, "wb") as f:
            pickle.dump(dummy_docs, f)

        # Load the docs back from file.
        with open(tmp_path, "rb") as f:
            loaded_docs = pickle.load(f)

        # Assert the loaded docs are identical to original
        assert len(loaded_docs) == len(dummy_docs)
        assert all([orig_doc == loaded_doc for orig_doc, loaded_doc in zip(dummy_docs, loaded_docs)])

        # Test that comparing Doc with int raises NotImplementedError
        with pytest.raises(NotImplementedError):
            loaded_docs[0] == 42



"""Allows chunking of documents into segments."""

import itertools
import re
import sys
from collections.abc import Callable, Iterable
from typing import Any

from sieves.data.doc import Doc
from sieves.tasks.core import Task


class NaiveChunker(Task):
    """Chunks by sentence counts. Only for test purposes."""

    def __init__(
        self,
        interval: int,
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize chunker.

        :param interval: Token count interval for chunks.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)
        self._interval = interval

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Split documents into chunks.

        :param docs: Documents to split.
        :return: Split documents.
        """
        batch_size = self._batch_size if self._batch_size > 0 else sys.maxsize
        while docs_batch := [doc for doc in itertools.islice(docs, batch_size)]:
            if len(docs_batch) == 0:
                break

            for doc in docs_batch:
                assert doc.text
                sentences = [sent for sent in re.split("[?!.]", doc.text) if len(sent.strip())]
                doc.chunks = [
                    ".".join(sentences[i : i + self._interval]) for i in range(0, len(sentences), self._interval)
                ]

                yield doc

    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "interval": self._interval,
        }



# mypy: ignore-errors
import pytest
from flaky import flaky

from sieves import Doc, Pipeline
from sieves.model_wrappers import ModelType, ModelSettings, dspy_, langchain_, outlines_
from sieves.serialization import Config
from sieves.tasks import PredictiveTask, QuestionAnswering
from sieves.tasks.predictive import question_answering
from sieves.tasks.predictive.schemas.question_answering import Result, QuestionAnswer


@flaky(max_runs=3, min_passes=1)
@pytest.mark.parametrize(
    "batch_runtime",
    QuestionAnswering.supports(),
    indirect=["batch_runtime"],
)
@pytest.mark.parametrize("fewshot", [True, False])
def test_run(qa_docs, batch_runtime, fewshot):
    fewshot_examples = [
        question_answering.FewshotExample(
            text="""
            Physics is the scientific study of matter, its fundamental constituents, its motion and behavior through
            space and time, and the related entities of energy and force. Physics is one of the most fundamental
            scientific disciplines. A scientist who specializes in the field of physics is called a physicist.
            """,
            questions=["What's a scientist called who specializes in the field of physics?"],
            answers=["A physicist."],
            scores=[1.0],
        ),
        question_answering.FewshotExample(
            text="""
            A biologist is a scientist who conducts research in biology. Biologists are interested in studying life on
            Earth, whether it is an individual cell, a multicellular organism, or a community of interacting
            populations. They usually specialize in a particular branch (e.g., molecular biology, zoology, and
            evolutionary biology) of biology and have a specific research focus (e.g., studying malaria or cancer).
            """,
            questions=["What are biologists interested in?"],
            answers=["Studying life."],
            scores=[1.0],
        ),
    ]

    fewshot_args = {"fewshot_examples": fewshot_examples} if fewshot else {}
    task = question_answering.QuestionAnswering(
        task_id="qa",
        questions=[
            "What branch of science is this text describing?",
            "What the goal of the science as described in the text?",
        ],
        model=batch_runtime.model,
        model_settings=batch_runtime.model_settings,
        batch_size=batch_runtime.batch_size,
        **fewshot_args,
    )
    pipe = Pipeline(task)
    docs = list(pipe(qa_docs))

    assert len(docs) == 2
    for doc in docs:
        assert doc.text
        assert "qa" in doc.results

        # Verify unified result types.
        assert isinstance(doc.results["qa"], question_answering.Result)

        assert "qa" in doc.meta
        assert "raw" in doc.meta["qa"]
        assert "usage" in doc.meta["qa"]
        assert "usage" in doc.meta

        print(f"Output: {doc.results['qa']}")
        print(f"Raw output: {doc.meta['qa']['raw']}")
        print(f"Usage: {doc.meta['qa']['usage']}")
        print(f"Total Usage: {doc.meta['usage']}")

    with pytest.raises(NotImplementedError):
        pipe["qa"].distill(None, None, None, None, None, None, None, None)

    if fewshot:
        _to_hf_dataset(task, docs)

def _to_hf_dataset(task: QuestionAnswering, docs: list[Doc]) -> None:
    """Tests whether conversion to HF dataset works as expected.

    :param task: QuestionAnswering task instance.
    :param docs: List of documents to convert.
    """
    dataset = task.to_hf_dataset(docs)
    assert all([key in dataset.features for key in ("text", "answers", "scores")])
    assert len(dataset) == 2
    dataset_records = list(dataset)
    for rec in dataset_records:
        assert isinstance(rec["text"], str)
        assert isinstance(rec["answers"], list)
        assert isinstance(rec["scores"], list)

    with pytest.raises(KeyError):
        task.to_hf_dataset([Doc(text="This is a dummy text.")])


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_serialization(qa_docs, batch_runtime) -> None:
    pipe = Pipeline(
        [
            question_answering.QuestionAnswering(
                task_id="qa",
                questions=[
                    "What branch of science is this text describing?",
                    "What the goal of the science as described in the text?",
                ],
                model=batch_runtime.model,
                model_settings=batch_runtime.model_settings,
                batch_size=batch_runtime.batch_size,
            )
        ]
    )

    config = pipe.serialize()
    assert config.model_dump() == {'cls_name': 'sieves.pipeline.core.Pipeline',
                                   'tasks': {'is_placeholder': False,
                                             'value': [{
                                                           'cls_name': 'sieves.tasks.predictive.question_answering.core.QuestionAnswering',
                                                           'fewshot_examples': {'is_placeholder': False,
                                                                                'value': ()},
                                                           'batch_size': {'is_placeholder': False, "value": -1},
                                                           'model_settings': {'is_placeholder': False,
                                                                                   'value': {
                                                                                       'config_kwargs': None,
                                                                                       'inference_kwargs': None,
                                                                                       'init_kwargs': None,
                                                                                       'strict': True,
                                                                                       'inference_mode': None}},
                                                           'include_meta': {'is_placeholder': False, 'value': True},
                                                           'model': {'is_placeholder': True,
                                                                     'value': 'dspy.clients.lm.LM'},
                                                           'prompt_instructions': {'is_placeholder': False,
                                                                                   'value': None},
                                                           'questions': {'is_placeholder': False,
                                                                         'value': ['What branch of science is this '
                                                                                   'text describing?',
                                                                                   'What the goal of the science as '
                                                                                   'described in the text?']},
                                                           'task_id': {'is_placeholder': False, 'value': 'qa'},
                                                           'condition': {'is_placeholder': False, 'value': None},
                                                           'version': Config.get_version()}]},
                                   'use_cache': {'is_placeholder': False, 'value': True},
                                   'version': Config.get_version()}

    Pipeline.deserialize(config=config, tasks_kwargs=[{"model": batch_runtime.model}])


@pytest.mark.parametrize(
    "batch_runtime",
    QuestionAnswering.supports(),
    indirect=["batch_runtime"],
)
def test_inference_mode_override(batch_runtime) -> None:
    """Test that inference_mode parameter overrides the default value."""
    dummy = "dummy_inference_mode"

    task = question_answering.QuestionAnswering(
        task_id="qa",
        questions=[
            "What branch of science is this text describing?",
            "What the goal of the science as described in the text?",
        ],
        model=batch_runtime.model,
        model_settings=ModelSettings(inference_mode=dummy),
        batch_size=batch_runtime.batch_size,
    )

    assert task._bridge.inference_mode == dummy


@pytest.mark.parametrize("batch_runtime", [ModelType.dspy], indirect=["batch_runtime"])
def test_evaluation(batch_runtime) -> None:
    """Test evaluation for QA using a real judge."""
    task = question_answering.QuestionAnswering(
        questions=["What is 1+1?"], model=batch_runtime.model, task_id="qa"
    )

    # 1. Full overlap
    doc_full = Doc(text="1+1 equals 2")
    res_full = Result(qa_pairs=[QuestionAnswer(question="What is 1+1?", answer="2")])
    doc_full.results["qa"] = res_full
    doc_full.gold["qa"] = res_full
    report_full = task.evaluate([doc_full], judge=batch_runtime.model)
    assert report_full.metrics[task.metric] > 0.8

    # 2. No overlap
    doc_none = Doc(text="1+1 equals 2")
    doc_none.results["qa"] = Result(qa_pairs=[QuestionAnswer(question="What is 1+1?", answer="It is a sunny day.")])
    doc_none.gold["qa"] = res_full
    report_none = task.evaluate([doc_none], judge=batch_runtime.model)
    assert report_none.metrics[task.metric] < 0.6

    # 3. Partial overlap (one correct, one incorrect)
    task_multi = question_answering.QuestionAnswering(
        questions=["What is 1+1?", "What is 2+2?"], model=batch_runtime.model, task_id="qa_multi"
    )
    doc_partial = Doc(text="1+1 is 2 and 2+2 is 4")
    # Pred: 2 and 5
    doc_partial.results["qa_multi"] = Result(qa_pairs=[
        QuestionAnswer(question="What is 1+1?", answer="2"),
        QuestionAnswer(question="What is 2+2?", answer="5")
    ])
    # Gold: 2 and 4
    doc_partial.gold["qa_multi"] = Result(qa_pairs=[
        QuestionAnswer(question="What is 1+1?", answer="2"),
        QuestionAnswer(question="What is 2+2?", answer="4")
    ])
    report_partial = task_multi.evaluate([doc_partial], judge=batch_runtime.model)
    # Expected to be somewhere in the middle
    assert 0.2 < report_partial.metrics[task_multi.metric] < 0.8



"""Question Answering predictive task."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, override

import datasets
import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import ModelType
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.question_answering.bridges import DSPyQuestionAnswering, PydanticQA
from sieves.tasks.predictive.schemas.question_answering import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)

_TaskBridge = DSPyQuestionAnswering | PydanticQA


class QuestionAnswering(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Answer questions about a text using structured model wrappers."""

    def __init__(
        self,
        questions: list[str],
        model: TaskModel,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """
        Initialize QuestionAnswering task.

        :param questions: Questions to answer.
        :param model: Model to use.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param model_settings: Settings for structured generation.
        :param condition: Optional callable that determines whether to process each document.
        """
        self._questions = questions
        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        return FewshotExample

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        for gold in truths:
            if gold is not None:
                assert isinstance(gold, TaskResult)
        for pred in preds:
            if pred is not None:
                assert isinstance(pred, TaskResult)

        return super()._compute_metrics(truths, preds, judge=judge)

    def _validate_fewshot_examples(self) -> None:
        """Validate that questions, answers and scores have the same length.

        :raises ValueError: If lengths don't match.
        """
        for i, example in enumerate(self._fewshot_examples):
            if len(example.questions) != len(example.answers):
                raise ValueError(
                    f"Length mismatch in few-shot example {i}: {len(example.questions)} questions "
                    f"vs {len(example.answers)} answers."
                )
            if example.scores is not None and len(example.scores) != len(example.answers):
                raise ValueError(
                    f"Length mismatch in few-shot example {i}: {len(example.answers)} answers "
                    f"vs {len(example.scores)} scores."
                )

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        bridge_types: dict[ModelType, type[_TaskBridge]] = {
            ModelType.dspy: DSPyQuestionAnswering,
            ModelType.outlines: PydanticQA,
            ModelType.langchain: PydanticQA,
        }

        try:
            return bridge_types[model_type](
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                questions=self._questions,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.langchain,
            ModelType.outlines,
        }

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        return TaskResult

    @override
    @property
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "questions": self._questions,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        # Define metadata.
        features = datasets.Features(
            {
                "text": datasets.Value("string"),
                "answers": datasets.Sequence(datasets.Value("string")),
                "scores": datasets.Sequence(datasets.Value("float32")),
            }
        )
        info = datasets.DatasetInfo(
            description=f"Question-answering dataset with questions {self._questions}. Generated with sieves "
            f"v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        try:
            data: list[tuple[str, list[str], list[float]]] = []
            for doc in docs:
                result = doc.results[self._task_id]
                answers = [qa.answer for qa in result.qa_pairs]
                scores = [qa.score if qa.score is not None else 1.0 for qa in result.qa_pairs]
                data.append((doc.text, answers, scores))
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        def generate_data() -> Iterable[dict[str, Any]]:
            """Yield results as dicts.

            :return: Results as dicts.
            """
            for text, answers, scores in data:
                yield {"text": text, "answers": answers, "scores": scores}

        # Create dataset.
        return datasets.Dataset.from_generator(generate_data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError



"""Marker task for converting PDF documents to text."""
# ruff: noqa: E402

import warnings
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any

# `marker` is using outdated Pydantic config approach, suppress these warnings.
warnings.filterwarnings(
    "ignore",
    message="Support for class-based `config` is deprecated.*ConfigDict instead",
    category=DeprecationWarning,
)

from marker.converters.pdf import PdfConverter
from marker.converters.table import TableConverter
from marker.models import create_model_dict
from marker.output import text_from_rendered

from sieves.data import Doc
from sieves.tasks.core import Task

Converter = PdfConverter | TableConverter


class Marker(Task):
    """Marker task for converting PDF documents to text."""

    def __init__(
        self,
        converter: Converter | None = None,
        export_format: str = "markdown",
        task_id: str | None = None,
        include_meta: bool = False,
        batch_size: int = -1,
        extract_images: bool = False,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize the Marker task.

        :param converter: Custom PdfConverter or TableConverter instance. If None, a default one will be created.
        :param export_format: Format to export the document in ("markdown", "html", or "json").
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for processing. Use -1 to process all documents at once.
        :param extract_images: Whether to extract images from the PDF.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(task_id=task_id, include_meta=include_meta, batch_size=batch_size, condition=condition)

        self._export_format = export_format
        self._converter = self._setup_converter(converter, self._export_format)
        self._extract_images = extract_images

    def _setup_converter(
        self, converter: PdfConverter | TableConverter | None, export_format: str
    ) -> PdfConverter | TableConverter:
        """Set up the converter with the specified renderer.

        :param converter: Custom converter instance or None.
        :param export_format: Format to export the document in.
        :return: Configured converter instance.
        """
        renderer: str = self._get_renderer(export_format)
        if converter is None:
            return PdfConverter(artifact_dict=create_model_dict(), renderer=renderer)

        # If a converter is provided, use its type but update the renderer
        if isinstance(converter, TableConverter):
            return TableConverter(artifact_dict=create_model_dict(), renderer=renderer)
        elif isinstance(converter, PdfConverter):
            return PdfConverter(artifact_dict=create_model_dict(), renderer=renderer)
        else:
            raise ValueError(f"Invalid converter type: {type(converter)}")

    def _get_renderer(self, export_format: str) -> str:
        """Get the renderer string based on the export format.

        :param export_format: Format to export the document in.
        :return: The renderer string.
        :raises ValueError: If the export format is invalid.
        """
        if export_format == "markdown":
            return "marker.renderers.markdown.MarkdownRenderer"
        elif export_format == "html":
            return "marker.renderers.html.HTMLRenderer"
        elif export_format == "json":
            return "marker.renderers.json.JSONRenderer"
        else:
            raise ValueError(f"Invalid export format: {export_format}")

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Process documents using Marker.

        :param docs: Documents to process.
        :return: Processed documents.
        """
        docs = list(docs)

        for doc in docs:
            # Convert URI to string if it's a Path
            uri = str(doc.uri) if isinstance(doc.uri, Path) else doc.uri
            # Process the document
            rendered = self._converter(uri)

            # Extract text and optionally images
            text, _, images = text_from_rendered(rendered)
            if self._extract_images:
                doc.images = images

            # Update document text
            doc.text = text

        for doc in docs:
            yield doc

    @property
    def _state(self) -> dict[str, Any]:
        """Get state for serialization.

        :return: State dictionary.
        """
        return {
            **super()._state,
            "converter": self._converter,
            "export_format": self._export_format,
            "extract_images": self._extract_images,
        }



"""Bridges for sentiment analysis task."""

import abc
from collections.abc import Callable, Sequence
from typing import Any, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.consolidation import MapScoreConsolidation
from sieves.tasks.predictive.schemas.sentiment_analysis import Result

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class SentimentAnalysisBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for sentiment analysis bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        aspects: tuple[str, ...],
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize sentiment analysis bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param aspects: Aspects to analyze.
        :param model_settings: Settings for structured generation.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        self._aspects = aspects
        self._consolidation_strategy = MapScoreConsolidation(extractor=self._chunk_extractor, keys=list(self._aspects))

    @property
    @abc.abstractmethod
    def _chunk_extractor(self) -> Callable[[Any], tuple[dict[str, float], float | None]]:
        """Return a callable that extracts (map_scores, overall_score) from a raw chunk result.

        :return: Extractor callable.
        """


class DSPySentimentAnalysis(SentimentAnalysisBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for sentiment analysis."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], tuple[dict[str, float], float | None]]:
        def extractor(res: Any) -> tuple[dict[str, float], float | None]:
            m_scores = {aspect: float(getattr(res, aspect)) for aspect in self._aspects}
            return m_scores, getattr(res, "score", None)

        return extractor

    @override
    def integrate(self, results: Sequence[dspy_.Result], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            # Take the first completion.
            prediction = result.completions[0]
            label_scores = {aspect: float(getattr(prediction, aspect)) for aspect in self._aspects}
            doc.results[self._task_id] = Result(
                sentiment_per_aspect=label_scores,
                score=getattr(prediction, "score", None),
            )
        return docs

    @override
    def consolidate(
        self, results: Sequence[dspy_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[dspy_.Result]:
        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)

        # Wrap back into dspy.Prediction.
        consolidated_results: list[dspy_.Result] = []
        for aspect_scores, overall_score in consolidated_results_clean:
            data = {**aspect_scores, "score": overall_score}
            consolidated_results.append(
                dspy.Prediction.from_completions(
                    [data],
                    signature=self.prompt_signature,
                )
            )
        return consolidated_results


class PydanticSentimentAnalysis(
    SentimentAnalysisBridge[pydantic.BaseModel, pydantic.BaseModel, ModelWrapperInferenceMode]
):
    """Pydantic-based sentiment analysis bridge."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        aspects_str = ",".join(self._aspects)
        return (
            f"Perform aspect-based sentiment analysis of the provided text given the provided aspects: {aspects_str}.\n"
            "For each aspect, provide the sentiment in the provided text with respect to this aspect.\n"
            "The 'overall' aspect should reflect the sentiment in the text overall.\n"
            "A score of 1.0 means that the sentiment in the text with respect to this aspect is extremely positive.\n"
            "0 means the opposite, 0.5 means neutral.\n"
            "The sentiment score per aspect should ALWAYS be between 0 and 1.\n"
            "Also provide an overall confidence score between 0.0 and 1.0 for the sentiment analysis."
        )

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return "========\n\n<text>{{ text }}</text>"

    @property
    @override
    def _chunk_extractor(self) -> Callable[[Any], tuple[dict[str, float], float | None]]:
        def extractor(res: Any) -> tuple[dict[str, float], float | None]:
            m_scores = {aspect: float(getattr(res, aspect)) for aspect in self._aspects}
            return m_scores, getattr(res, "score", None)

        return extractor

    @override
    def integrate(self, results: Sequence[pydantic.BaseModel], docs: list[Doc]) -> list[Doc]:
        for doc, result in zip(docs, results):
            label_scores = {k: v for k, v in result.model_dump().items() if k not in ["score"]}
            doc.results[self._task_id] = Result(sentiment_per_aspect=label_scores, score=getattr(result, "score", None))
        return docs

    @override
    def consolidate(
        self, results: Sequence[pydantic.BaseModel], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[pydantic.BaseModel]:
        assert issubclass(self.prompt_signature, pydantic.BaseModel)

        consolidated_results_clean = self._consolidation_strategy.consolidate(results, docs_offsets)
        consolidated_results: list[pydantic.BaseModel] = []

        for aspect_scores, overall_score in consolidated_results_clean:
            consolidated_results.append(
                self.prompt_signature(
                    **aspect_scores,
                    score=overall_score,
                )
            )

        return consolidated_results

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json
        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



"""
Tests for task usage documentation snippets.
"""
from typing import Any
import pytest
import pydantic
import dspy
from sieves import Pipeline, tasks, Doc
from sieves.tasks.predictive.information_extraction import FewshotExampleMulti, FewshotExampleSingle

# Use a real model class but with dummy credentials.
# This satisfies type checks and initialization logic without making network calls.
model = dspy.LM("openai/gpt-4o-mini", api_key="dummy")

def test_classification_usage():
    # --8<-- [start:classification-list]
    from sieves import tasks

    task = tasks.Classification(
        labels=["positive", "negative"],
        model=model,
    )
    # --8<-- [end:classification-list]
    assert task

    # --8<-- [start:classification-dict]
    task = tasks.Classification(
        labels={
            "positive": "Positive sentiment regarding the subject.",
            "negative": "Negative sentiment regarding the subject.",
        },
        model=model,
    )
    # --8<-- [end:classification-dict]
    assert task

def test_ner_usage():
    # --8<-- [start:ner-usage]
    from sieves import tasks

    task = tasks.NER(
        entities=["PERSON", "ORGANIZATION", "LOCATION"],
        model=model,
    )
    # --8<-- [end:ner-usage]
    assert task

    # --8<-- [start:ner-dict-usage]
    task = tasks.NER(
        entities={
            "PERSON": "Names of people.",
            "ORGANIZATION": "Companies, agencies, institutions.",
        },
        model=model,
    )
    # --8<-- [end:ner-dict-usage]
    assert task

def test_information_extraction_usage():
    # --8<-- [start:ie-multi]
    import pydantic
    from sieves import tasks
    from sieves.tasks.predictive.information_extraction import FewshotExampleMulti

    class Person(pydantic.BaseModel, frozen=True):
        name: str
        age: int

    examples = [
        FewshotExampleMulti(
            text="Alice is 30 and Bob is 25.",
            entities=[Person(name="Alice", age=30), Person(name="Bob", age=25)]
        )
    ]

    task = tasks.InformationExtraction(
        entity_type=Person,
        mode="multi",
        fewshot_examples=examples,
        model=model,
    )
    # --8<-- [end:ie-multi]
    assert task

    # --8<-- [start:ie-single]
    from sieves import tasks
    from sieves.tasks.predictive.information_extraction import FewshotExampleSingle

    class Invoice(pydantic.BaseModel, frozen=True):
        id: str
        total: float

    examples = [
        FewshotExampleSingle(
            text="Invoice #123: $50.00",
            entity=Invoice(id="123", total=50.0)
        )
    ]

    task = tasks.InformationExtraction(
        entity_type=Invoice,
        mode="single",
        fewshot_examples=examples,
        model=model,
    )
    # --8<-- [end:ie-single]
    assert task

def test_sentiment_analysis_usage():
    # --8<-- [start:sentiment-usage]
    from sieves import tasks

    task = tasks.SentimentAnalysis(
        model=model,
    )
    # --8<-- [end:sentiment-usage]
    assert task

def test_question_answering_usage():
    # --8<-- [start:qa-usage]
    from sieves import tasks

    task = tasks.QuestionAnswering(
        questions=["What is the main topic?", "Who are the key figures?"],
        model=model,
    )
    # --8<-- [end:qa-usage]
    assert task

def test_summarization_usage():
    # --8<-- [start:summarization-usage]
    from sieves import tasks

    task = tasks.Summarization(
        model=model,
        n_words=100, # Guideline for summary length
    )
    # --8<-- [end:summarization-usage]
    assert task

def test_translation_usage():
    # --8<-- [start:translation-usage]
    from sieves import tasks

    task = tasks.Translation(
        to="Spanish",
        model=model,
    )
    # --8<-- [end:translation-usage]
    assert task

def test_pii_masking_usage():
    # --8<-- [start:pii-usage]
    from sieves import tasks

    task = tasks.PIIMasking(
        model=model,
    )
    # --8<-- [end:pii-usage]
    assert task

def test_chunking_usage():
    # --8<-- [start:chunking-usage]
    import chonkie
    from sieves import tasks

    # Define a chunker (e.g., using Chonkie)
    chunker_model = chonkie.RecursiveChunker()

    task = tasks.Chunking(
        chunker=chunker_model,
    )
    # --8<-- [end:chunking-usage]
    assert task

def test_ingestion_usage():
    # --8<-- [start:ingestion-usage]
    from sieves import tasks

    task = tasks.Ingestion(
        export_format="markdown",
    )
    # --8<-- [end:ingestion-usage]
    assert task



"""Chat Control demo for Marimo."""
import marimo

__generated_with = "0.18.3"
app = marimo.App(
    width="full",
    layout_file="layouts/demo_chatcontrol.slides.json",
)

with app.setup:
    # Initialization code that runs before all other cells
    pass


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # OSA Community Demo: Sieves

    2025-12-12
    """)
    return


@app.cell
def _():
    import marimo as mo

    return (mo,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Chat Control is a hot topic in European politics. In order to get up to date on current developments, we'll analyze a recent article of the Electronic Frontier Foundation on this topic.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Building a dataset

    We define a single document referencing a recent EFF article on this:
    """)
    return


@app.cell
def _():
    from sieves import Doc

    docs = [
        Doc(
            uri="https://www.eff.org/deeplinks/2025/12/after-years-controversy-eus-chat-control-nears-its-final-hurdle-what-know"
        )
    ]
    return (docs,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Before we run anything, we'll need to define our model. In this case we go with DSPy (which uses LiteLLM underneath), but other libraries
    are supported:
    - Hugging Face zero-shot pipelines
    - GliNER2
    - Outlines
    - LangChain
    """)
    return


@app.cell
def _():
    import os

    import dspy

    openrouter_api_base = "https://openrouter.ai/api/v1/"
    openrouter_model_id = "anthropic/claude-haiku-4.5"

    model = dspy.LM(
        f"openrouter/{openrouter_model_id}", api_base=openrouter_api_base, api_key=os.environ["OPENROUTER_API_KEY"]
    )
    return (model,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## Questions...
    - What's the summary of the article?
    - What are possible risks of introducing ChatControl?
    - Why is ChatControl an issue _again_ after being rejected already?
    - How positive or negative is the author's stance on chat control?
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Let's define a pipeline that reflects our requirements. As first task we add an _ingestion_ task to convert the document into plain text, then we define the tasks that address our questions.
    """)
    return


@app.cell
def _(model):
    from sieves import tasks

    pipe = (
        tasks.Ingestion()
        + tasks.QuestionAnswering(
            questions=[
                "What are the risks of introducing ChatControl?",
                "I thought ChatControl was rejected. Why is this an issue again?",
            ],
            model=model,
        )
        + tasks.Summarization(n_words=20, model=model)
        + tasks.SentimentAnalysis(aspects=["digital rights organisation", "chat control"], model=model)
    )
    return pipe, tasks


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## ...(changing our mind)...
    Hold on - maybe we want to read further literature on this?

    To find related articles, we first must define what we're looking for, so that the model can pick up on it.
    """)
    return


@app.cell
def _():
    import pydantic

    class RelatedLiterature(pydantic.BaseModel, frozen=True):
        """Another article on the topic of ChatControl."""

        topic: str = pydantic.Field(description="Topic or title of the article on this.")
        link: str = pydantic.Field(description="Link to the article.")

    return (RelatedLiterature,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Now we can extend our pipeline:
    """)
    return


@app.cell
def _(RelatedLiterature, model, pipe, tasks):
    full_pipe = pipe + tasks.InformationExtraction(entity_type=RelatedLiterature, model=model)
    return (full_pipe,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    ## ...and answers!

    We're finally ready to extract results:
    """)
    return


@app.cell
def _(docs, full_pipe):
    doc_with_results = [result for result in full_pipe(docs)][0]
    result = doc_with_results.results
    return (result,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    **What's the summary of the article?**
    """)
    return


@app.cell
def _(result):
    from pprint import pprint

    pprint(result["Summarization"])
    return (pprint,)


@app.cell
def _(mo):
    mo.md(r"""
    **What are possible risks of introducing ChatControl?**
    """)
    return


@app.cell
def _(pprint, result):
    pprint(result["QuestionAnswering"][0])
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    **Why is ChatControl an issue again after being rejected already?**
    """)
    return


@app.cell
def _(pprint, result):
    pprint(result["QuestionAnswering"][1])
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    **How positive or negative is the author's stance on chat control?**
    """)
    return


@app.cell(hide_code=True)
def _(pprint, result):
    pprint(result["SentimentAnalysis"])
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    **Further EFF literature on this topic:**
    """)
    return


@app.cell
def _(result):
    for article in result["InformationExtraction"]:
        print(article.topic)
        print(f"{article.link}")
        print("#######")
    return


@app.cell
def _():
    return


if __name__ == "__main__":
    app.run()



"""Bridge base class and types."""

from __future__ import annotations

import abc
import inspect
import warnings
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import pydantic

with warnings.catch_warnings():
    warnings.simplefilter("ignore", category=DeprecationWarning)
    from pydantic_ai.format_as_xml import format_as_xml

from sieves.data import Doc
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.utils import convert_to_signature

if TYPE_CHECKING:
    from sieves.model_wrappers import ModelType


class Bridge[TaskPromptSignature, TaskResult, ModelWrapperInferenceMode](abc.ABC):
    """Bridge base class."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        overwrite: bool,
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize new bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param overwrite: Whether to overwrite text with produced text. Considered only by bridges for tasks producing
            fluent text - like translation, summarization, PII masking, etc.
        :param model_settings: Model settings including inference_mode.
        :param prompt_signature: Pydantic model class representing the task's output schema.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        self._task_id = task_id
        self._custom_prompt_instructions = prompt_instructions
        self._overwrite = overwrite
        self._model_settings = model_settings
        self._pydantic_signature = prompt_signature
        self._model_type = model_type
        self._fewshot_examples = fewshot_examples

        self._validate()

    def _validate(self) -> None:
        """Validate configuration.

        No-op by default. Executed at the end of __init__().
        """

    @property
    @abc.abstractmethod
    def _default_prompt_instructions(self) -> str:
        """Return default prompt instructions.

        Instructions are injected at the beginning of each prompt.

        :return: Default prompt instructions.
        """

    @property
    def _prompt_instructions(self) -> str:
        """Returns prompt instructions.

        :returns: If `_custom_prompt_instructions` is set, this is used. Otherwise, `_default_prompt_instructions` is
            used.
        """
        return self._custom_prompt_instructions or self._default_prompt_instructions

    @property
    def _prompt_example_xml(self) -> str | None:
        """Return prompt template for example injection.

        Examples are injected between instructions and conclusions.

        :return: Default prompt example template.
        """
        if not self._fewshot_examples:
            return None

        # format_as_xml handles escaping and structured formatting.
        # Passing a list of models usually results in an <examples> root tag.
        return format_as_xml(self._fewshot_examples).strip()

    @property
    def _prompt_conclusion(self) -> str | None:
        """Return prompt conclusion.

        Prompt conclusions are injected at the end of each prompt.

        :return: Default prompt conclusion.
        """
        return None

    @property
    def model_settings(self) -> ModelSettings:
        """Return model settings.

        :return: Model settings.
        """
        return self._model_settings

    @property
    def model_type(self) -> ModelType:
        """Return model type.

        :return: Model type.
        """
        return self._model_type

    @property
    def prompt_template(self) -> str:
        """Return prompt template.

        Chains `_prompt_instructions`, `_prompt_example_xml` and `_prompt_conclusion`.

        Note: different model have different expectations as to how a prompt should look like. E.g. outlines supports
        the Jinja 2 templating format for insertion of values and few-shot examples, whereas DSPy integrates these
        things in a different value in the workflow and hence expects the prompt not to include these things. Mind
        model-specific expectations when creating a prompt template.
        :return str | None: Prompt template as string. None if not used by model wrapper.
        """
        instructions = inspect.cleandoc(self._custom_prompt_instructions or self._prompt_instructions)
        examples = (self._prompt_example_xml or "").strip()
        conclusion = inspect.cleandoc(self._prompt_conclusion or "")

        prompt_parts = [instructions]
        if examples:
            prompt_parts.append(examples)
        if conclusion:
            prompt_parts.append(conclusion)

        return "\n\n".join(prompt_parts).strip()

    @property
    def prompt_signature(self) -> type[TaskPromptSignature] | TaskPromptSignature:
        """Create output signature.

        E.g.: `Signature` in DSPy, Pydantic objects in outlines, JSON schema in jsonformers.
        This is model type-specific.

        :return type[_TaskPromptSignature] | _TaskPromptSignature: Output signature object. This can be an instance
            (e.g. a regex string) or a class (e.g. a Pydantic class).
        """
        # Extract framework-specific kwargs if needed.
        kwargs: dict[str, Any] = {}
        if self.model_settings.inference_mode:
            kwargs["inference_mode"] = self.model_settings.inference_mode

        return convert_to_signature(  # type: ignore[invalid-return-type]
            model_cls=self._pydantic_signature,
            model_type=self.model_type,
            **kwargs,
        )

    @property
    @abc.abstractmethod
    def inference_mode(self) -> ModelWrapperInferenceMode:
        """Return inference mode.

        :return ModelWrapperInferenceMode: Inference mode.
        """

    def extract(self, docs: Sequence[Doc]) -> Sequence[dict[str, Any]]:
        """Extract all values from doc instances that are to be injected into the prompts.

        :param docs: Docs to extract values from.
        :return: All values from doc instances that are to be injected into the prompts as a sequence.
        """
        return [{"text": doc.text if doc.text else None} for doc in docs]

    @abc.abstractmethod
    def integrate(self, results: Sequence[TaskResult], docs: list[Doc]) -> list[Doc]:
        """Integrate results into Doc instances.

        :param results: Results from prompt executable.
        :param docs: Doc instances to update.
        :return: Updated doc instances as a list.
        """

    @abc.abstractmethod
    def consolidate(self, results: Sequence[TaskResult], docs_offsets: list[tuple[int, int]]) -> Sequence[TaskResult]:
        """Consolidate results for document chunks into document results.

        :param results: Results per document chunk.
        :param docs_offsets: Chunk offsets per document. Chunks per document can be obtained with
            `results[docs_chunk_offsets[i][0]:docs_chunk_offsets[i][1]]`.
        :return: Results per document as a sequence.
        """



# mypy: ignore-errors
from typing import Literal, Union, Any

import dspy
import gliner2.inference.engine
import pydantic
import pytest

from sieves.model_wrappers import ModelType
from sieves.tasks.predictive.utils import convert_to_signature

# --- Test Models ---

class SimpleExtraction(pydantic.BaseModel):
    """Simple extraction model."""
    name: str = pydantic.Field(description="The name of the person")
    age: int = pydantic.Field(description="The age of the person")

class ClassificationModel(pydantic.BaseModel):
    """Classification model."""
    label: Literal["A", "B", "C"] = pydantic.Field(description="The category")
    score: float

class MultiLabelModel(pydantic.BaseModel):
    """Multi-label model."""
    science: bool
    politics: bool
    sports: bool

# --- Helpers for comparison ---

def assert_dspy_sigs_equal(actual: type[dspy.Signature], expected: type[dspy.Signature]):
    """Deeply compare two DSPy signatures."""
    assert actual.input_fields.keys() == expected.input_fields.keys()
    assert actual.output_fields.keys() == expected.output_fields.keys()

    for k in actual.input_fields:
        assert actual.input_fields[k].json_schema_extra.get("desc") == expected.input_fields[k].json_schema_extra.get("desc")

    for k in actual.output_fields:
        assert actual.output_fields[k].json_schema_extra.get("desc") == expected.output_fields[k].json_schema_extra.get("desc")

    assert actual.__doc__ == expected.__doc__

def assert_gliner_schemas_equal(actual: Any, expected: Any):
    """Compare two GliNER schemas by their underlying dictionary representation."""
    if hasattr(actual, "_field_metadata"):
        assert actual._field_metadata == expected._field_metadata
    elif hasattr(actual, "schema"):
        if hasattr(actual.schema, "_field_metadata"):
            assert actual.schema._field_metadata == expected.schema._field_metadata
        else:
            assert actual.schema == expected.schema
    else:
        assert actual == expected

# --- Tests ---

def test_convert_to_dspy_full():
    # Define manually what we expect
    class ExpectedSig(dspy.Signature):
        """Simple extraction model."""
        text = dspy.InputField(desc="Input text to process.")
        name = dspy.OutputField(desc="The name of the person")
        age = dspy.OutputField(desc="The age of the person")

    actual = convert_to_signature(SimpleExtraction, ModelType.dspy)

    assert_dspy_sigs_equal(actual, ExpectedSig)

def test_convert_to_gliner_structure_full():
    # Expected manually constructed GliNER2 structure
    expected = gliner2.inference.engine.Schema().structure("SimpleExtraction")
    expected.field("name", dtype="str")
    expected.field("age", dtype="str")

    actual = convert_to_signature(SimpleExtraction, ModelType.gliner, mode="structure")

    assert_gliner_schemas_equal(actual, expected)

def test_convert_to_gliner_classification_full():
    # Expected manually constructed GliNER2 classification
    expected = gliner2.inference.engine.Schema().classification(
        task="classification",
        labels=["A", "B", "C"]
    )

    actual = convert_to_signature(ClassificationModel, ModelType.gliner, mode="classification")

    assert_gliner_schemas_equal(actual, expected)

def test_convert_to_huggingface_full():
    # Instruction: grab names of all fields that are not called 'score'.
    expected = ["A", "B", "C"]
    actual = convert_to_signature(ClassificationModel, ModelType.huggingface)
    assert actual == expected

    expected_multi = ["science", "politics", "sports"]
    actual_multi = convert_to_signature(MultiLabelModel, ModelType.huggingface)
    assert actual_multi == expected_multi

def test_convert_to_outlines_choice_full():
    # For choice mode, it should extract labels from Literal
    expected = ["A", "B", "C"]
    actual = convert_to_signature(ClassificationModel, ModelType.outlines, inference_mode="choice")
    assert actual == expected

def test_convert_to_gliner_entities_full():
    expected = gliner2.inference.engine.Schema().entities(
        entity_types=["science", "politics", "sports"]
    )

    actual = convert_to_signature(MultiLabelModel, ModelType.gliner, mode="entities")

    assert_gliner_schemas_equal(actual, expected)

def test_convert_to_gliner_structure_choices_full():
    class ChoiceModel(pydantic.BaseModel):
        category: Literal["X", "Y"]

    expected = gliner2.inference.engine.Schema().structure("ChoiceModel")
    expected.field("category", dtype="str", choices=["X", "Y"])

    actual = convert_to_signature(ChoiceModel, ModelType.gliner, mode="structure")

    assert_gliner_schemas_equal(actual, expected)



"""Outlines model_wrapper."""

import asyncio
import enum
from collections.abc import Iterable, Sequence
from typing import Any, Literal, override

import json_repair
import outlines
import pydantic
import tiktoken
from outlines.models import AsyncBlackBoxModel, BlackBoxModel, SteerableModel

from sieves.model_wrappers.core import Executable, PydanticModelWrapper
from sieves.model_wrappers.types import TokenUsage

PromptSignature = (
    pydantic.BaseModel | list[str] | str | outlines.types.Choice | outlines.types.Regex | outlines.types.JsonSchema
)
Model = AsyncBlackBoxModel | BlackBoxModel | SteerableModel
Result = pydantic.BaseModel | str


class InferenceMode(enum.Enum):
    """Available inference modes.

    Note: generator functions are wrapped in tuples, as otherwise the Enum instance seems to be replaced by the function
    itself - not sure why that happens. Should take another look at this.
    """

    # For normal text output, i.e. no structured generation.
    text = "text"
    # For limited set of choices, e.g. classification.
    choice = "choice"
    # Regex-conforming output.
    regex = "regex"
    # Output conforming to Pydantic models.
    json = "json"


class Outlines(PydanticModelWrapper[PromptSignature, Result, Model, InferenceMode]):
    """ModelWrapper for Outlines with multiple structured inference modes."""

    @override
    @property
    def inference_modes(self) -> type[InferenceMode]:
        return InferenceMode

    async def _generate_async(
        self,
        generator: (
            outlines.generator.SteerableGenerator
            | outlines.generator.BlackBoxGenerator
            | outlines.generator.AsyncBlackBoxGenerator
        ),
        prompt: str,
    ) -> Result | None:
        """Generate result async.

        :param generator: Generator instance to use for generation.
        :param prompt: Prompt to generate result for.
        :return: Result for prompt. Results are None if corresponding prompt failed.
        """
        result = generator(prompt, **self._inference_kwargs)
        assert isinstance(result, Result) or result is None
        return result

    @override
    def build_executable(
        self,
        inference_mode: InferenceMode,
        prompt_template: str | None,  # noqa: UP007
        prompt_signature: type[PromptSignature] | PromptSignature,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ) -> Executable[Result | None]:
        # Set a moderate default for max_new_tokens for `transformers` models, as they otherwise run into truncation
        # issues.
        inference_kwargs = self._inference_kwargs.copy()
        if isinstance(self._model, outlines.models.transformers.Transformers):
            inference_kwargs = {"max_new_tokens": 1024} | inference_kwargs

        template = self._create_template(prompt_template)

        # Create Generator instance responsible for generating non-parsed text.
        if isinstance(prompt_signature, list):
            prompt_signature = Literal[*prompt_signature]  # type: ignore[invalid-type-form]

        if inference_mode == InferenceMode.regex:
            prompt_signature = outlines.types.Regex(prompt_signature)

        generator = outlines.Generator(self._model, output_type=prompt_signature, **self._init_kwargs)

        def execute(values: Sequence[dict[str, Any]]) -> Sequence[tuple[Result | None, Any, TokenUsage]]:
            """Execute prompts with model wrapper for given values.

            :param values: Values to inject into prompts.
            :return: Sequence of tuples containing results, raw outputs, and token usage. Results are None if
                corresponding prompt failed.
            """

            def generate(prompts: list[str]) -> Iterable[tuple[Result, Any, TokenUsage]]:
                try:
                    results = generator.batch(prompts, **inference_kwargs)
                # Batch mode is not implemented for all Outlines wrappers. Fall back to single-prompt mode in
                # that case.
                except NotImplementedError:
                    calls = [self._generate_async(generator, prompt) for prompt in prompts]
                    results = asyncio.run(self._execute_async_calls(calls))

                # Estimate token usage if tokenizer is available.
                tokenizer = self._get_tokenizer()

                if inference_mode == InferenceMode.json:
                    assert len(results) == len(prompts)
                    assert isinstance(prompt_signature, type) and issubclass(prompt_signature, pydantic.BaseModel)

                    for prompt, result in zip(prompts, results):
                        usage = TokenUsage(
                            input_tokens=self._count_tokens(prompt, tokenizer),
                            output_tokens=self._count_tokens(result, tokenizer),
                        )

                        try:
                            parsed = prompt_signature.model_validate_json(result)
                            yield parsed, result, usage
                        # If naive parsing fails: JSON is potentially invalid. Attempt to repair it, then try again.
                        except pydantic.ValidationError:
                            repaired = json_repair.repair_json(result)
                            parsed = prompt_signature.model_validate_json(repaired)
                            yield parsed, result, usage

                else:
                    for prompt, result in zip(prompts, results):
                        usage = TokenUsage(
                            input_tokens=self._count_tokens(prompt, tokenizer),
                            output_tokens=self._count_tokens(str(result), tokenizer),
                        )

                        yield result, result, usage

            return self._infer(generate, template, values)

        return execute

    @override
    def _get_tokenizer(self) -> Any | None:
        # Outlines models usually have a tokenizer, but some wrappers (like OpenAI) might hide it.
        tokenizer = getattr(self._model, "tokenizer", None)
        if not tokenizer and hasattr(self._model, "model"):
            tokenizer = getattr(self._model.model, "tokenizer", None)

        # Fall back to tiktoken for remote OpenAI-compatible models if no tokenizer found.
        if not tokenizer:
            try:
                # Default to o200k_base for newer models if specific model encoding not found.
                tokenizer = tiktoken.get_encoding("o200k_base")
            except ImportError:
                tokenizer = None

        return tokenizer



"""Aspect-based sentiment analysis predictive task."""

from __future__ import annotations

from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, override

import datasets
import dspy
import pydantic
import sklearn

from sieves.data import Doc
from sieves.model_wrappers import ModelType
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.schemas.sentiment_analysis import (
    FewshotExample,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)
from sieves.tasks.predictive.sentiment_analysis.bridges import (
    DSPySentimentAnalysis,
    PydanticSentimentAnalysis,
)

_TaskBridge = DSPySentimentAnalysis | PydanticSentimentAnalysis


class SentimentAnalysis(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Estimate per‑aspect and overall sentiment for a document."""

    THRESHOLD: float = 0.5

    def __init__(
        self,
        model: TaskModel,
        model_settings: ModelSettings = ModelSettings(),
        aspects: tuple[str, ...] = tuple(),
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """
        Initialize SentimentAnalysis task.

        :param model: Model to use.
        :param model_settings: Settings for structured generation.
        :param aspects: Aspects to consider in sentiment analysis. Overall sentiment will always be determined. If
            empty, only overall sentiment will be determined.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param condition: Optional callable that determines whether to process each document.
        """
        self._aspects = tuple(sorted(set(aspects) | {"overall"}))
        super().__init__(
            model=model,
            model_settings=model_settings,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        return FewshotExample

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        """Return the unified Pydantic prompt signature for this task.

        :return: Unified Pydantic prompt signature.
        """
        fields = {
            aspect: (
                float,
                pydantic.Field(
                    ...,
                    description=f"Sentiment score for the '{aspect}' aspect, ranging from 0 (Negative) to 1 "
                    f"(Positive).",
                    ge=0,
                    le=1,
                ),
            )
            for aspect in self._aspects
        }
        fields["score"] = (
            float | None,
            pydantic.Field(
                default=None,
                description="Provide an overall confidence score for the sentiment analysis, ranging from 0 to 1.",
            ),
        )

        return pydantic.create_model(
            "SentimentAnalysisOutput",
            __base__=pydantic.BaseModel,
            __doc__="Result of aspect-based sentiment analysis.",
            **fields,
        )  # type: ignore[no-matching-overload]

    @property
    @override
    def metric(self) -> str:
        return "F1 (Macro)"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        Computes F1 per aspect, then averages over aspects. Sentiments are discretized to 0 (Negative) or 1 (Positive)
        for F1.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        aspects_list = list(self._aspects)
        # We will compute F1 per aspect, then average.
        # Sentiments are 0.0 to 1.0. We discretize them to 0 (Negative) or 1 (Positive) for F1.

        scores: list[float] = []

        for aspect in aspects_list:
            y_true: list[int] = []
            y_pred: list[int] = []

            for gold, pred in zip(truths, preds):
                if gold is not None:
                    assert isinstance(gold, TaskResult)
                    gold_val = gold.sentiment_per_aspect.get(aspect, 0.0)
                    y_true.append(1 if gold_val >= self.THRESHOLD else 0)
                else:
                    y_true.append(-1)

                if pred is not None:
                    assert isinstance(pred, TaskResult)
                    pred_val = pred.sentiment_per_aspect.get(aspect, 0.0)
                    y_pred.append(1 if pred_val >= self.THRESHOLD else 0)
                else:
                    y_pred.append(-1)

            # Compute Macro F1 for this aspect.
            if y_true:
                scores.append(sklearn.metrics.f1_score(y_true, y_pred, average="macro", zero_division=0))

        avg_score = sum(scores) / len(scores) if scores else 0.0

        return {self.metric: avg_score}

    @override
    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        bridge_types: dict[ModelType, type[DSPySentimentAnalysis | PydanticSentimentAnalysis]] = {
            ModelType.dspy: DSPySentimentAnalysis,
            ModelType.outlines: PydanticSentimentAnalysis,
            ModelType.langchain: PydanticSentimentAnalysis,
        }

        try:
            bridge_type = bridge_types[model_type]

            return bridge_type(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                aspects=self._aspects,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.langchain,
            ModelType.outlines,
        }

    @override
    def _validate_fewshot_examples(self) -> None:
        for fs_example in self._fewshot_examples:
            assert isinstance(fs_example, FewshotExample)
            if any([aspect not in self._aspects for aspect in fs_example.sentiment_per_aspect]) or not all(
                [label in fs_example.sentiment_per_aspect for label in self._aspects]
            ):
                raise ValueError(
                    f"Aspect mismatch: {self._task_id} has aspects {self._aspects}. Few-shot examples have "
                    f"aspects {fs_example.sentiment_per_aspect.keys()}."
                )

    @property
    @override
    def _state(self) -> dict[str, Any]:
        return {
            **super()._state,
            "aspects": self._aspects,
        }

    @override
    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        threshold = threshold or self.THRESHOLD

        # Define metadata.
        features = datasets.Features(
            {
                "text": datasets.Value("string"),
                "aspect": datasets.Sequence(datasets.Value("float32")),
                "score": datasets.Value("float32"),
            }
        )
        info = datasets.DatasetInfo(
            description=f"Aspect-based sentiment analysis dataset with aspects {self._aspects}. Generated with sieves "
            f"v{Config.get_version()}.",
            features=features,
        )

        # Fetch data used for generating dataset.
        aspects = self._aspects
        try:
            data = [
                (
                    doc.text,
                    doc.results[self._task_id].sentiment_per_aspect,
                    doc.results[self._task_id].score,
                )
                for doc in docs
            ]
        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        def generate_data() -> Iterable[dict[str, Any]]:
            """Yield results as dicts.

            :return: Results as dicts.
            """
            for text, scores, score in data:
                yield {"text": text, "aspect": [scores[aspect] for aspect in aspects], "score": score}

        # Create dataset.
        return datasets.Dataset.from_generator(generate_data, features=features, info=info)

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        raise NotImplementedError

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        # Compute per-aspect accuracy as 1 - abs(true_sentiment - pred_sentiment)
        # Average across all aspects (same approach as multi-label classification)
        accuracy = 0
        for aspect, sentiment in truth["sentiment_per_aspect"].items():
            if aspect in pred["sentiment_per_aspect"]:
                pred_sentiment = max(min(pred["sentiment_per_aspect"][aspect], 1), 0)
                accuracy += 1 - abs(sentiment - pred_sentiment)

        base_accuracy = accuracy / len(truth["sentiment_per_aspect"])

        # If score is available in both truth and pred, incorporate it into the metric.
        if "score" in truth and truth["score"] is not None and "score" in pred and pred["score"] is not None:
            score_accuracy = 1 - abs(truth["score"] - max(min(pred["score"], 1), 0))
            return (base_accuracy + score_accuracy) / 2

        return base_accuracy



"""Relation extraction task."""

from sieves.tasks.predictive.relation_extraction.core import RelationExtraction
from sieves.tasks.predictive.schemas.relation_extraction import (
    FewshotExample,
    RelationEntity,
    RelationTriplet,
    Result,
)

__all__ = ["RelationExtraction", "FewshotExample", "RelationEntity", "RelationTriplet", "Result"]



"""Tests for conditional task execution."""

from __future__ import annotations

from collections.abc import Callable, Iterable

from sieves.data import Doc
from sieves.pipeline import Pipeline
from sieves.tasks.core import Task


class DummyTask(Task):
    """Simple dummy task for testing conditional execution."""

    def __init__(
        self,
        task_id: str | None = None,
        condition: Callable[[Doc], bool] | None = None,
    ):
        """Initialize DummyTask.

        :param task_id: Task ID.
        :param condition: Optional callable that determines whether to process each document.
        """
        super().__init__(
            task_id=task_id or "DummyTask",
            include_meta=False,
            batch_size=-1,
            condition=condition,
        )

    def _call(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """Process documents by marking them as processed.

        :param docs: Documents to process.
        :return: Processed documents.
        """
        for doc in docs:
            # Mark document as processed
            doc.results[self.id] = {"processed": True}
            yield doc


def test_condition_none_processes_all_docs() -> None:
    """Test that when condition is None, all documents are processed."""
    docs = [
        Doc(text="Short"),
        Doc(text="This is a longer text"),
        Doc(text="Medium length text"),
    ]

    task = DummyTask(condition=None)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 3
    # All docs should have been processed
    assert all(doc.results[task.id] is not None for doc in result)


def test_condition_true_for_all_processes_all_docs() -> None:
    """Test that when condition returns True for all docs, all are processed."""
    docs = [
        Doc(text="Short"),
        Doc(text="This is a longer text"),
        Doc(text="Medium length text"),
    ]

    def always_true(doc: Doc) -> bool:
        return True

    task = DummyTask(condition=always_true)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 3
    assert all(doc.results[task.id] is not None for doc in result)


def test_condition_false_for_all_skips_all_docs() -> None:
    """Test that when condition returns False for all docs, none are processed."""
    docs = [
        Doc(text="Short"),
        Doc(text="This is a longer text"),
        Doc(text="Medium length text"),
    ]

    task = DummyTask(condition=lambda d: False)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 3
    # All docs should have None in results (indicating they were skipped by the condition)
    for doc in result:
        assert doc.results[task.id] is None


def test_condition_mixed_processing() -> None:
    """Test that documents are selectively processed based on condition."""
    docs = [
        Doc(text="Short"),
        Doc(text="This is a much longer text that should be processed"),
        Doc(text="Medium"),
        Doc(text="Another long document that exceeds the short threshold here"),
    ]

    task = DummyTask(condition=lambda d: len(d.text or "") > 20)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 4

    # First doc is short - should be skipped
    assert result[0].results[task.id] is None

    # Second doc is long - should be processed
    assert result[1].results[task.id] is not None

    # Third doc is short - should be skipped
    assert result[2].results[task.id] is None

    # Fourth doc is long - should be processed
    assert result[3].results[task.id] is not None


def test_condition_preserves_document_order() -> None:
    """Test that document order is preserved with conditional execution."""
    docs = [Doc(text=f"Doc {i}") for i in range(5)]

    def is_even_index(doc: Doc) -> bool:
        try:
            idx = int(doc.text.split()[-1])
            return idx % 2 == 0
        except (ValueError, IndexError):
            return False

    task = DummyTask(condition=is_even_index)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    # Order should be maintained
    assert len(result) == 5
    for i, doc in enumerate(result):
        assert doc.text == f"Doc {i}"
        assert doc.results['DummyTask'] != i % 2


def test_condition_in_pipeline() -> None:
    """Test that conditions work correctly in multi-task pipelines."""
    docs = [
        Doc(text="short"),
        Doc(text="this is a longer document that should be processed"),
        Doc(text="med"),
    ]

    def is_long(doc: Doc) -> bool:
        return len(doc.text or "") > 15

    # Create two tasks with different conditions
    task1 = DummyTask(task_id="task1", condition=lambda d: len(d.text or "") > 10)
    task2 = DummyTask(task_id="task2", condition=is_long)

    pipe = Pipeline([task1, task2])
    result = list(pipe(docs))

    assert len(result) == 3

    # First doc: short - skipped by both
    assert result[0].results.get("task1") is None
    assert result[0].results.get("task2") is None

    # Second doc: long - processed by both
    assert result[1].results.get("task1") is not None
    assert result[1].results.get("task2") is not None

    # Third doc: medium - skipped by both
    assert result[2].results.get("task1") is None
    assert result[2].results.get("task2") is None


def test_condition_none_result_storage() -> None:
    """Test that skipped documents have None in results."""
    docs = [Doc(text="Test"), Doc(text="Skip this one")]

    def skip_second(doc: Doc) -> bool:
        return doc.text != "Skip this one"

    task = DummyTask(condition=skip_second)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    # First doc processed
    assert result[0].results[task.id] is not None
    # Second doc skipped
    assert result[1].results[task.id] is None

    # But order is preserved
    assert result[0].text == "Test"
    assert result[1].text == "Skip this one"


def test_condition_exception_handling() -> None:
    """Test that exceptions in condition functions are properly raised."""
    docs = [Doc(text="Test")]

    def bad_condition(doc: Doc) -> bool:
        raise ValueError("Intentional error")

    task = DummyTask(condition=bad_condition)
    pipe = Pipeline([task])

    try:
        list(pipe(docs))
        assert False, "Should have raised ValueError"
    except ValueError as e:
        assert "Intentional error" in str(e)


def test_condition_with_task_id() -> None:
    """Test that conditions work with custom task IDs."""
    docs = [
        Doc(text="short"),
        Doc(text="this is a much longer text that will be processed"),
    ]

    def is_long(doc: Doc) -> bool:
        return len(doc.text or "") > 20

    # Use task with custom ID and condition
    task = DummyTask(task_id="custom_task", condition=is_long)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 2

    # First doc is short - skipped
    assert result[0].results["custom_task"] is None

    # Second doc is long - processed
    assert result[1].results["custom_task"] is not None


def test_multiple_docs_same_text_with_condition() -> None:
    """Test handling of duplicate docs (same text) with conditions."""
    docs = [
        Doc(text="Duplicate text"),
        Doc(text="Duplicate text"),  # Same text
        Doc(text="Different text"),
    ]

    def skip_duplicates(doc: Doc) -> bool:
        return doc.text != "Duplicate text"

    task = DummyTask(condition=skip_duplicates)
    pipe = Pipeline([task])
    result = list(pipe(docs))

    assert len(result) == 3

    # First two docs skipped (duplicates)
    assert result[0].results[task.id] is None
    assert result[1].results[task.id] is None

    # Third doc processed
    assert result[2].results[task.id] is not None



"""GLiNER2 model wrapper wrapper built on top of GLiNER2 multi‑task pipelines."""

import enum
import warnings
from collections.abc import Sequence
from typing import Any, override

import gliner2
import jinja2
import pydantic

from sieves.model_wrappers.core import Executable, ModelWrapper
from sieves.model_wrappers.types import TokenUsage

PromptSignature = gliner2.inference.engine.Schema | gliner2.inference.engine.StructureBuilder
Model = gliner2.GLiNER2
Result = dict[str, str | list[str | dict[str, Any]]]


class InferenceMode(enum.Enum):
    """Available inference modes."""

    classification = 1
    entities = 2
    structure = 3
    relations = 4


class GliNER(ModelWrapper[PromptSignature, Result, Model, InferenceMode]):
    """ModelWrapper adapter for GLiNER2."""

    @override
    @property
    def inference_modes(self) -> type[InferenceMode]:
        return InferenceMode

    @override
    @property
    def supports_few_shotting(self) -> bool:
        return False

    @override
    def build_executable(
        self,
        inference_mode: InferenceMode,
        prompt_template: str | None,
        prompt_signature: type[PromptSignature] | PromptSignature,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ) -> Executable[Result]:
        cls_name = self.__class__.__name__
        if len(list(fewshot_examples)):
            warnings.warn(f"Few-shot examples are not supported by model type {cls_name}.")

        # Overwrite prompt default template, if template specified. Note that this is a static prompt and GliNER doesn't
        # do few-shotting, so we don't inject anything into the template.
        if prompt_template:
            self._model.prompt = jinja2.Template(prompt_template).render()

        def execute(values: Sequence[dict[str, Any]]) -> Sequence[tuple[Result, Any, TokenUsage]]:
            """Execute prompts with model wrapper for given values.

            :param values: Values to inject into prompts.
            :return: Sequence of tuples containing results, raw outputs, and token usage.
            """
            results = self._model.batch_extract(
                texts=[val["text"] for val in values],
                schemas=prompt_signature,
                **(
                    {"batch_size": len(values)}
                    | self._inference_kwargs
                    | {"include_confidence": True, "include_spans": True}
                ),
            )

            # Estimate token usage if tokenizer is available.
            tokenizer = self._get_tokenizer()

            final_results: list[tuple[Result, Any, TokenUsage]] = []
            # Estimate token counts.
            for val, res in zip(values, results):
                usage = TokenUsage(
                    input_tokens=self._count_tokens(val["text"], tokenizer),
                    output_tokens=self._count_tokens(str(res), tokenizer),
                )

                final_results.append((res, res, usage))
            return final_results

        return execute

    @override
    def _get_tokenizer(self) -> Any | None:
        # GLiNER2 models usually have a tokenizer accessible through the processor.
        tokenizer = getattr(self._model, "tokenizer", None)
        if not tokenizer and hasattr(self._model, "processor"):
            tokenizer = getattr(self._model.processor, "tokenizer", None)

        return tokenizer



"""
Tests for Doc usage documentation snippets.
"""
from sieves import Doc

def test_doc_usage():
    # --8<-- [start:doc-usage]
    from sieves import Doc

    # Create a document from text
    doc = Doc(text="This is a sample document.")

    # Create a document with metadata
    doc_with_meta = Doc(
        text="Document with metadata.",
        meta={"source": "manual", "priority": "high"}
    )

    # Documents can also have a URI
    doc_with_uri = Doc(uri="https://example.com/doc.pdf")
    # --8<-- [end:doc-usage]
    assert doc.text == "This is a sample document."



"""Information extraction task."""

from sieves.tasks.predictive.information_extraction.core import InformationExtraction
from sieves.tasks.predictive.schemas.information_extraction import (
    FewshotExampleMulti,
    FewshotExampleSingle,
    ResultMulti,
    ResultSingle,
)

__all__ = ["InformationExtraction", "FewshotExampleMulti", "FewshotExampleSingle", "ResultMulti", "ResultSingle"]



"""Bridges for NER task."""

import abc
import re
from collections.abc import Sequence
from typing import Any, TypeVar, override

import dspy
import pydantic

from sieves.data import Doc
from sieves.model_wrappers import (
    ModelType,
    ModelWrapperInferenceMode,
    dspy_,
    langchain_,
    outlines_,
)
from sieves.model_wrappers.types import ModelSettings
from sieves.tasks.predictive.bridges import Bridge
from sieves.tasks.predictive.schemas.ner import Entity, Result
from sieves.tasks.predictive.utils import convert_to_signature

_BridgePromptSignature = TypeVar("_BridgePromptSignature")
_BridgeResult = TypeVar("_BridgeResult")


class NERBridge(Bridge[_BridgePromptSignature, _BridgeResult, ModelWrapperInferenceMode], abc.ABC):
    """Abstract base class for NER bridges."""

    def __init__(
        self,
        task_id: str,
        prompt_instructions: str | None,
        entities: list[str] | dict[str, str],
        model_settings: ModelSettings,
        prompt_signature: type[pydantic.BaseModel],
        model_type: ModelType,
        fewshot_examples: Sequence[pydantic.BaseModel] = (),
    ):
        """Initialize NER bridge.

        :param task_id: Task ID.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param entities: List of entities to extract or dict mapping labels to descriptions.
        :param model_settings: Settings for structured generation.
        :param prompt_signature: Unified Pydantic prompt signature.
        :param model_type: Model type.
        :param fewshot_examples: Few-shot examples.
        """
        assert model_type in {ModelType.dspy, ModelType.outlines, ModelType.langchain, ModelType.gliner}

        super().__init__(
            task_id=task_id,
            prompt_instructions=prompt_instructions,
            overwrite=False,
            model_settings=model_settings,
            prompt_signature=prompt_signature,
            model_type=model_type,
            fewshot_examples=fewshot_examples,
        )
        if isinstance(entities, dict):
            self._entities = list(entities.keys())
            self._entity_descriptions = entities
        else:
            self._entities = list(entities)
            self._entity_descriptions = {}

    @override
    @property
    def prompt_signature(self) -> _BridgePromptSignature:
        return convert_to_signature(
            model_cls=self._pydantic_signature,
            model_type=self.model_type,
            mode="entities",
        )  # type: ignore[return-value]

    def _get_entity_descriptions(self) -> str:
        """Return a string with the entity descriptions.

        :return: A string with the entity descriptions.
        """
        entities_with_descriptions: list[str] = []
        for entity in self._entities:
            if entity in self._entity_descriptions:
                entities_with_descriptions.append(
                    f"  <entity_description>\n    <entity>{entity}</entity>\n    <description>"
                    f"{self._entity_descriptions[entity]}</description>\n  </entity_description>"
                )
            else:
                entities_with_descriptions.append(f"  <entity>{entity}</entity>")

        entity_desc_string = "\n".join(entities_with_descriptions)
        return f"<entity_descriptions>\n{entity_desc_string}\n</entity_descriptions>"

    @staticmethod
    def _find_entity_positions(
        doc_text: str,
        result: _BridgeResult,
    ) -> list[Entity]:
        """Find all positions of an entity in a document.

        :param doc_text: The text of the document.
        :param result: The result of the model.
        :return: The list of entities with start/end indices.
        """
        doc_text_lower = doc_text.lower()
        # Create a new result with the same structure as the original
        new_entities: list[Entity] = []

        # Track entities by position to avoid duplicates
        entities_by_position: dict[tuple[int, int], Entity] = {}
        context_list: list[str] = []

        entities_list = getattr(result, "entities", [])
        for entity_with_context in entities_list:
            # Skip if there is no entity
            if not entity_with_context:
                continue

            # Get the entity and context texts from the model
            entity_text = getattr(entity_with_context, "text", "")
            context = getattr(entity_with_context, "context", None)
            entity_type = getattr(entity_with_context, "entity_type", "")
            score = getattr(entity_with_context, "score", None)

            if not entity_text:
                continue

            if context is None:
                new_entities.append(
                    Entity(
                        text=entity_text,
                        start=-1,
                        end=-1,
                        entity_type=entity_type,
                        score=score,
                    )
                )
                continue

            entity_text_lower = entity_text.lower()
            context_lower = context.lower() if context else ""
            # Create a list of the unique contexts
            # Avoid adding duplicates as entities witht he same context would be captured twice
            if context_lower not in context_list:
                context_list.append(context_lower)
            else:
                continue
            # Find all occurrences of the context in the document using regex
            context_positions = re.finditer(re.escape(context_lower), doc_text_lower)

            # For each context position that was found (usually is just one), find the entity within that context
            for match in context_positions:
                context_start = match.start()
                entity_start_in_context = context_lower.find(entity_text_lower)

                if entity_start_in_context >= 0:
                    start = context_start + entity_start_in_context
                    end = start + len(entity_text)

                    # Create a new entity with start/end indices
                    new_entity = Entity(
                        text=doc_text[start:end],
                        start=start,
                        end=end,
                        entity_type=entity_type,
                        score=score,
                    )

                    # Only add if this exact position hasn't been filled yet
                    position_key = (start, end)
                    if position_key not in entities_by_position:
                        entities_by_position[position_key] = new_entity
                        new_entities.append(new_entity)

        return sorted(new_entities, key=lambda x: x.start)

    @override
    def integrate(self, results: Sequence[_BridgeResult], docs: list[Doc]) -> list[Doc]:
        docs_list = list(docs)
        results_list = list(results)

        for doc, result in zip(docs_list, results_list):
            # Get the original text from the document
            doc_text = doc.text or ""
            if hasattr(result, "entities"):
                # Process entities from result if available
                entities_with_position = self._find_entity_positions(doc_text, result)
                # Create a new result with the updated entities
                new_result = Result(text=doc_text, entities=entities_with_position)
                doc.results[self._task_id] = new_result
            else:
                # Default empty result
                doc.results[self._task_id] = Result(text=doc_text, entities=[])

        return docs_list


class DSPyNER(NERBridge[dspy_.PromptSignature, dspy_.Result, dspy_.InferenceMode]):
    """DSPy bridge for NER."""

    @override
    def _validate(self) -> None:
        assert self._model_type == ModelType.dspy

    @override
    @property
    def model_type(self) -> ModelType:
        return ModelType.dspy

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        return ""

    @override
    @property
    def inference_mode(self) -> dspy_.InferenceMode:
        return self._model_settings.inference_mode or dspy_.InferenceMode.predict

    @override
    def consolidate(
        self, results: Sequence[dspy_.Result], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[dspy_.Result]:
        # Process each document (which may consist of multiple chunks)
        consolidated_results: list[dspy_.Result] = []
        for doc_offset in docs_offsets:
            doc_results = results[doc_offset[0] : doc_offset[1]]

            # Combine all entities from all chunks
            all_entities: list[Entity] = []

            # Process each chunk for this document
            for chunk_result in doc_results:
                if chunk_result is None:
                    continue

                if not hasattr(chunk_result, "entities") or not chunk_result.entities:
                    continue

                # Process entities in this chunk
                for entity in chunk_result.entities:
                    all_entities.append(entity)

            # Create a consolidated result for this document
            consolidated_results.append(
                dspy.Prediction.from_completions({"entities": [all_entities]}, signature=self.prompt_signature)
            )
        return consolidated_results


class PydanticNER(NERBridge[pydantic.BaseModel, pydantic.BaseModel, ModelWrapperInferenceMode], abc.ABC):
    """Base class for Pydantic-based NER bridges."""

    @override
    def _validate(self) -> None:
        assert self._model_type in {ModelType.langchain, ModelType.outlines}

    @override
    @property
    def _default_prompt_instructions(self) -> str:
        entity_info = self._get_entity_descriptions() if self._entity_descriptions else ""
        return (
            "Your goal is to extract named entities from the text. Only extract entities of the specified types:\n"
            f"{self._entities}.\n"
            f"{entity_info}\n\n"
            "For each entity:\n"
            "- Extract the exact text of the entity\n"
            "- Include a SHORT context string that contains ONLY the entity and AT MOST 3 words before and 3 words "
            "after it.\n"
            "  DO NOT include the entire text as context. DO NOT include words that are not present in the original "
            "text\n"
            "  as introductory words (Eg. 'Text:' before context string).\n"
            "- Specify which type of entity it is (must be one of the provided entity types)\n"
            "- Provide a confidence score between 0.0 and 1.0 for the extraction.\n\n"
            "IMPORTANT:\n"
            "- If the same entity appears multiple times in the text, extract each occurrence separately with its own "
            "context"
        )

    @override
    @property
    def _prompt_conclusion(self) -> str | None:
        return "===========\n\n<text>{{ text }}</text>\n<entity_types>{{ entity_types }}</entity_types>\n<entities>"

    @override
    def consolidate(
        self, results: Sequence[pydantic.BaseModel], docs_offsets: list[tuple[int, int]]
    ) -> Sequence[pydantic.BaseModel]:
        assert issubclass(self.prompt_signature, pydantic.BaseModel)

        # Process each document (which may consist of multiple chunks).
        consolidated_results: list[pydantic.BaseModel] = []
        for doc_offset in docs_offsets:
            doc_results = results[doc_offset[0] : doc_offset[1]]

            # Combine all entities from all chunks
            all_entities: list[dict[str, Any]] = []

            # Process each chunk for this document
            for chunk_result in doc_results:
                if chunk_result is None:
                    continue

                if not hasattr(chunk_result, "entities") or not chunk_result.entities:
                    continue

                # Process entities in this chunk
                for entity in chunk_result.entities:
                    # We just need to combine all entities from all chunks
                    all_entities.append(entity)

            # Create a consolidated result for this document - instantiate the class with entities
            consolidated_results.append(self.prompt_signature(entities=all_entities))

        return consolidated_results

    @override
    @property
    def model_type(self) -> ModelType:
        return self._model_type

    @override
    @property
    def inference_mode(self) -> outlines_.InferenceMode | langchain_.InferenceMode:
        if self._model_type == ModelType.outlines:
            return self._model_settings.inference_mode or outlines_.InferenceMode.json
        elif self._model_type == ModelType.langchain:
            return self._model_settings.inference_mode or langchain_.InferenceMode.structured

        raise ValueError(f"Unsupported model type: {self._model_type}")



"""Classification predictive task and few‑shot example schemas."""

from __future__ import annotations

import json
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, Literal, override

import datasets
import dspy
import pydantic
import sklearn

from sieves.data import Doc
from sieves.model_wrappers import ModelType, gliner_
from sieves.model_wrappers.types import ModelSettings
from sieves.serialization import Config
from sieves.tasks.distillation.distillation_import import model2vec, setfit
from sieves.tasks.distillation.types import DistillationFramework
from sieves.tasks.predictive.classification.bridges import (
    DSPyClassification,
    HuggingFaceClassification,
    LangChainClassification,
    OutlinesClassification,
)
from sieves.tasks.predictive.core import PredictiveTask
from sieves.tasks.predictive.gliner_bridge import GliNERBridge
from sieves.tasks.predictive.schemas.classification import (
    FewshotExampleMultiLabel,
    FewshotExampleSingleLabel,
    ResultMultiLabel,
    ResultSingleLabel,
    TaskModel,
    TaskPromptSignature,
    TaskResult,
)

_TaskBridge = (
    DSPyClassification | GliNERBridge | LangChainClassification | HuggingFaceClassification | OutlinesClassification
)

FewshotExample = FewshotExampleMultiLabel | FewshotExampleSingleLabel


class Classification(PredictiveTask[TaskPromptSignature, TaskResult, _TaskBridge]):
    """Predictive task for text classification."""

    def __init__(
        self,
        labels: Sequence[str] | dict[str, str],
        model: TaskModel,
        task_id: str | None = None,
        include_meta: bool = True,
        batch_size: int = -1,
        prompt_instructions: str | None = None,
        fewshot_examples: Sequence[FewshotExample] = (),
        mode: Literal["single", "multi"] = "multi",
        model_settings: ModelSettings = ModelSettings(),
        condition: Callable[[Doc], bool] | None = None,
    ) -> None:
        """Initialize new Classification task.

        :param labels: Labels to predict. Supports two formats:
            - List format: `["science", "politics", "sports"]`
            - Dict format: `{"science": "Scientific topics", "politics": "Political topics"}`
            The dict format allows you to provide descriptions that help the model better understand each label.
        :param model: Model to use.
        :param task_id: Task ID.
        :param include_meta: Whether to include meta information generated by the task.
        :param batch_size: Batch size to use for inference. Use -1 to process all documents at once.
        :param prompt_instructions: Custom prompt instructions. If None, default instructions are used.
        :param fewshot_examples: Few-shot examples.
        :param mode: If 'multi', task returns confidence scores for all specified labels. If 'single', task returns
            most likely class label.
        :param model_settings: Model settings.
        :param condition: Optional callable that determines whether to process each document.
        """
        if isinstance(labels, dict):
            self._labels = list(labels.keys())
            self._label_descriptions = labels
        else:
            self._labels = list(labels)
            self._label_descriptions = {}
        self._mode = mode

        super().__init__(
            model=model,
            task_id=task_id,
            include_meta=include_meta,
            batch_size=batch_size,
            overwrite=False,
            prompt_instructions=prompt_instructions,
            fewshot_examples=fewshot_examples,
            model_settings=model_settings,
            condition=condition,
        )

    @property
    @override
    def fewshot_example_type(self) -> type[FewshotExample]:
        """Return few-shot example type.

        :return: Few-shot example type.
        """
        if self._mode == "multi":
            return FewshotExampleMultiLabel

        return FewshotExampleSingleLabel

    @property
    @override
    def prompt_signature(self) -> type[pydantic.BaseModel]:
        if self._mode == "single":
            labels = self._labels
            LabelType = Literal[*labels]  # type: ignore[valid-type]

            class SingleLabelClassification(pydantic.BaseModel):
                """Result of single-label classification. Contains the most likely label and its confidence score."""

                label: LabelType = pydantic.Field(description="The predicted label from the specified set of labels.")
                score: float = pydantic.Field(
                    default=None, description="Provide a confidence score for the predicted label, between 0 and 1."
                )

            return SingleLabelClassification

        # For multi-label, create a model with fields for each label.
        fields: dict[str, tuple[type, Any]] = {}
        for label in self._labels:
            assert isinstance(self._label_descriptions, dict)
            description = self._label_descriptions.get(
                label, f"Confidence score for the '{label}' category, between 0 and 1."
            )  # type: ignore[no-matching-overload]
            fields[label] = (float, pydantic.Field(description=description))

        return pydantic.create_model(  # type: ignore[no-matching-overload]
            "MultiLabelClassification",
            __base__=pydantic.BaseModel,
            __doc__="Result of multi-label classification. Contains confidence scores for each potential label.",
            **fields,
        )

    @property
    @override
    def metric(self) -> str:
        return "F1 (Macro)"

    @override
    def _compute_metrics(self, truths: list[Any], preds: list[Any], judge: dspy.LM | None = None) -> dict[str, float]:
        """Compute corpus-level metrics.

        :param truths: List of ground truths.
        :param preds: List of predictions.
        :param judge: Optional DSPy LM instance to use as judge for generative tasks.
        :return: Dictionary of metrics.
        """
        # Prepare labels and mappings.
        labels_list = list(self._labels)
        label_to_idx = {label: i for i, label in enumerate(labels_list)}

        y_true: list[float | list[float] | int] = []
        y_pred: list[float | list[float] | int] = []

        for gold, pred in zip(truths, preds):
            if gold is None or pred is None:
                # If either is None, we handle it as a failure for this example.
                # In multi-label, this is an all-zero vector. In single-label, it's -1 (no class).
                if self._mode == "multi":
                    y_true.append([0.0] * len(labels_list))
                    y_pred.append([0.0] * len(labels_list))
                else:
                    y_true.append(-1)
                    y_pred.append(-1)
                continue

            # Convert to normalized representation using the existing logic.
            # We reuse `_task_result_to_dspy_dict` indirectly or just use `_result_to_scores`.
            gold_scores = self._result_to_scores(self._task_result_to_pydantic(gold))
            pred_scores = self._result_to_scores(self._task_result_to_pydantic(pred))

            if self._mode == "multi":
                # Binary multi-hot vectors.
                y_true.append([1.0 if gold_scores.get(label, 0.0) >= self.THRESHOLD else 0.0 for label in labels_list])
                y_pred.append([1.0 if pred_scores.get(label, 0.0) >= self.THRESHOLD else 0.0 for label in labels_list])
            else:
                # Single label indices.
                # Get the label with the highest score.
                gold_label = max(gold_scores.items(), key=lambda x: x[1])[0]
                pred_label = max(pred_scores.items(), key=lambda x: x[1])[0]
                y_true.append(label_to_idx.get(gold_label, -1))
                y_pred.append(label_to_idx.get(pred_label, -1))

        # Filter out examples where gold or pred were None (represented by -1 or all-zero if that's what we chose).
        score = sklearn.metrics.f1_score(y_true, y_pred, average="macro", zero_division=0)

        return {self.metric: float(score)}

    def _init_bridge(self, model_type: ModelType) -> _TaskBridge:
        """Initialize bridge.

        :return: ModelWrapper task.
        :raises ValueError: If model type is not supported.
        """
        # Reconstruct labels parameter (as dict if descriptions exist, else as list)
        labels = self._label_descriptions if self._label_descriptions else self._labels

        if model_type == ModelType.gliner:
            return GliNERBridge(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                prompt_signature=self.prompt_signature,
                model_settings=self._model_settings,
                inference_mode=gliner_.InferenceMode.classification,
                mode=self._mode,
            )

        bridge_types: dict[ModelType, type[_TaskBridge]] = {
            ModelType.dspy: DSPyClassification,
            ModelType.huggingface: HuggingFaceClassification,
            ModelType.outlines: OutlinesClassification,
            ModelType.langchain: LangChainClassification,
        }

        try:
            bridge_type = bridge_types[model_type]
            assert not issubclass(bridge_type, GliNERBridge)

            return bridge_type(
                task_id=self._task_id,
                prompt_instructions=self._custom_prompt_instructions,
                labels=labels,
                mode=self._mode,
                model_settings=self._model_settings,
                prompt_signature=self.prompt_signature,
                model_type=model_type,
                fewshot_examples=self._fewshot_examples,
            )
        except KeyError as err:
            raise KeyError(f"Model type {model_type} is not supported by {self.__class__.__name__}.") from err

    @staticmethod
    @override
    def supports() -> set[ModelType]:
        return {
            ModelType.dspy,
            ModelType.gliner,
            ModelType.huggingface,
            ModelType.langchain,
            ModelType.outlines,
        }

    def _validate_fewshot_examples(self) -> None:
        label_error_text = (
            "Label mismatch: {task_id} has labels {labels}. Few-shot examples have labels {example_labels}."
        )
        example_type_error_text = "Fewshot example type mismatch: mode = {mode} requires {example_type}."

        for fs_example in self._fewshot_examples or []:
            if self._mode == "multi":
                assert isinstance(fs_example, FewshotExampleMultiLabel), TypeError(
                    example_type_error_text.format(example_type=FewshotExampleMultiLabel, mode=self._mode)
                )
                if any([label not in self._labels for label in fs_example.score_per_label]) or not all(
                    [label in fs_example.score_per_label for label in self._labels]
                ):
                    raise ValueError(
                        label_error_text.format(
                            task_id=self.id, labels=self._labels, example_labels=fs_example.score_per_label.keys()
                        )
                    )
            else:
                assert isinstance(fs_example, FewshotExampleSingleLabel), TypeError(
                    example_type_error_text.format(example_type=FewshotExampleSingleLabel, mode=self._mode)
                )
                if fs_example.label not in self._labels:
                    raise ValueError(
                        label_error_text.format(task_id=self.id, labels=self._labels, example_labels=(fs_example.label))
                    )

    @property
    def _state(self) -> dict[str, Any]:
        # Store labels as dict if descriptions exist, otherwise as list
        labels = self._label_descriptions if self._label_descriptions else self._labels
        return {
            **super()._state,
            "labels": labels,
        }

    @staticmethod
    def _result_to_scores(result: ResultMultiLabel | ResultSingleLabel) -> dict[str, float]:
        """Normalize a single result to a mapping of label → score.

        :params result: One result value from ``doc.results``.

        :return: Mapping from label to score.

        :raises TypeError: If the result has an unsupported type or shape.

        """
        if isinstance(result, ResultMultiLabel):
            return {str(label): float(score) for label, score in result.label_scores}

        if isinstance(result, ResultSingleLabel):
            return {result.label: result.score}

        raise TypeError(f"Unsupported result type in _result_to_scores: {type(result)}")

    @override
    def distill(
        self,
        base_model_id: str,
        framework: DistillationFramework,
        data: datasets.Dataset | Sequence[Doc],
        output_path: Path | str,
        val_frac: float,
        init_kwargs: dict[str, Any] | None = None,
        train_kwargs: dict[str, Any] | None = None,
        seed: int | None = None,
    ) -> None:
        init_kwargs = init_kwargs or {}
        train_kwargs = train_kwargs or {}
        output_path = Path(output_path)
        output_path.mkdir(parents=True, exist_ok=True)

        data = self.to_hf_dataset(data) if isinstance(data, Sequence) else data

        required_columns = {"text", "labels"}
        if not required_columns.issubset(data.column_names):
            raise ValueError(f"Dataset must contain columns: {required_columns}. Found: {data.column_names}")

        dataset_splits = self._split_dataset(data, 1 - val_frac, val_frac, seed)
        dataset_splits.save_to_disk(output_path / "data")

        match framework:
            case DistillationFramework.setfit:
                default_init_kwargs: dict[str, Any] = {}
                metric_kwargs: dict[str, Any] = {}

                if self._mode == "multi":
                    default_init_kwargs["multi_target_strategy"] = "multi-output"
                    metric_kwargs = {"average": "macro"}

                model = setfit.SetFitModel.from_pretrained(base_model_id, **(default_init_kwargs | init_kwargs))

                args = setfit.TrainingArguments(
                    output_dir=str(output_path),
                    eval_strategy="epoch",
                    save_strategy="epoch",
                    load_best_model_at_end=True,
                    **train_kwargs,
                )

                trainer = setfit.Trainer(
                    model=model,
                    args=args,
                    train_dataset=dataset_splits["train"],
                    eval_dataset=dataset_splits.get("val"),
                    metric="f1",
                    column_mapping={"text": "text", "labels": "label"},
                    metric_kwargs=metric_kwargs,
                )
                trainer.train()
                trainer.model.save_pretrained(output_path)

                metrics = trainer.evaluate()
                with open(output_path / "metrics.json", "w") as f:
                    json.dump(metrics, f, indent=4)

            case DistillationFramework.model2vec:

                def one_hot_to_label(label_indices: list[int]) -> list[str]:
                    """Convert list of label indices into list of labels.

                    :param label_indices: List of label indices.
                    :return: List of labels.
                    """
                    return [str(self._labels[i]) for i, is_label in enumerate(label_indices) if is_label]

                classifier = model2vec.train.StaticModelForClassification.from_pretrained(
                    model_name=base_model_id, **init_kwargs
                )
                classifier.fit(
                    dataset_splits["train"]["text"],
                    [one_hot_to_label(encoded_labels) for encoded_labels in dataset_splits["train"]["labels"]],
                    **train_kwargs,
                )
                classifier.to_pipeline().save_pretrained(output_path)

                metrics = classifier.evaluate(
                    dataset_splits["val"]["text"],
                    [one_hot_to_label(encoded_labels) for encoded_labels in dataset_splits["val"]["labels"]],
                )
                with open(output_path / "metrics.json", "w") as f:
                    json.dump(metrics, f, indent=4)

            case _:
                raise NotImplementedError(
                    f"Unsupported distillation framework for this task: {framework}. "
                    f"Please choose one of {DistillationFramework.setfit, DistillationFramework.model2vec}"
                )

    def to_hf_dataset(self, docs: Iterable[Doc], threshold: float | None = None) -> datasets.Dataset:
        """Convert results to a Hugging Face dataset with multi-hot labels.

        The emitted dataset contains a ``text`` column and a ``labels`` column which is a multi-hot list aligned to
        ``self._labels``. This method is robust to different result shapes produced by various model wrappers and
        bridges in both single-label and multi-label configurations:
        - ``list[tuple[str, float]]`` for multi-label results
        - ``tuple[str, float]`` for single-label results
        - ``str`` for single-label results (assumes score ``1.0``)
        - ``pydantic.BaseModel`` exposing ``label`` and optional ``score``

        :param docs: Documents whose ``results`` contain outputs for this task id.
        :param threshold: Threshold to convert scores into multi-hot indicators. Defaults to `THRESHOLD`.

        :return: A ``datasets.Dataset`` with ``text`` and multi-hot ``labels``.

        :raises KeyError: If any document is missing this task's results.
        :raises TypeError: If a result cannot be interpreted.

        """
        threshold = threshold or self.THRESHOLD

        data: list[dict[str, str | list[bool]]] = []

        # Define metadata and features (multi-hot across declared labels for multi-label).
        if self._mode == "multi":
            features = datasets.Features(
                {"text": datasets.Value("string"), "labels": datasets.Sequence(datasets.Value("bool"))}
            )
        else:
            features = datasets.Features(
                {"text": datasets.Value("string"), "labels": datasets.ClassLabel(names=self._labels)}
            )

        info = datasets.DatasetInfo(
            description=(
                f"{'Multi-label' if self._mode == 'multi' else 'Single-label'} classification dataset with labels "
                f"{self._labels}. Generated with sieves v{Config.get_version()}."
            ),
            features=features,
        )

        try:
            for doc in docs:
                scores = Classification._result_to_scores(doc.results[self._task_id])

                # If multi-label: store one-hot representation.
                if self._mode == "multi":
                    result_normalized = [int(scores.get(label, 0.0) >= threshold) for label in self._labels]  # type: ignore[no-matching-overload]
                # If single-label: get single-label result as is.
                else:
                    keys = list(scores.keys())
                    assert len(keys) == 1
                    result_normalized = keys[0]

                data.append({"text": doc.text, "labels": result_normalized})

        except KeyError as err:
            raise KeyError(f"Not all documents have results for this task with ID {self._task_id}") from err

        return datasets.Dataset.from_list(data, features=features, info=info)

    def _task_result_to_pydantic(self, result: Any) -> ResultSingleLabel | ResultMultiLabel:
        """Convert a result to the appropriate Pydantic model for this task.

        :param result: Input result.
        :return: Normalized Pydantic result.
        """
        if isinstance(result, ResultSingleLabel | ResultMultiLabel):
            return result

        # Use the same logic as `_task_result_to_dspy_dict` but return the Pydantic model.
        if isinstance(result, str):
            return ResultSingleLabel(label=result, score=1.0)
        elif isinstance(result, list) and all(isinstance(item, str) for item in result):
            if self._mode == "single":
                if len(result) == 1:
                    return ResultSingleLabel(label=result[0], score=1.0)
                else:
                    raise ValueError(f"Got list of {len(result)} labels for single-label task.")
            else:
                label_scores: list[tuple[str, float]] = []
                active_set = set(result)
                for label in self._labels:
                    s = 1.0 if label in active_set else 0.0
                    label_scores.append((label, s))
                return ResultMultiLabel(label_scores=label_scores)

        raise TypeError(f"Unsupported result type in classification task: {type(result)}")

    @override
    def _task_result_to_dspy_dict(self, result: Any) -> dict[str, Any]:
        # We accept `str` and `list[str]` for better UX.
        if isinstance(result, str):
            result = ResultSingleLabel(label=result, score=1.0)
        elif isinstance(result, list) and all(isinstance(item, str) for item in result):
            if self._mode == "single":
                if len(result) == 1:
                    result = ResultSingleLabel(label=result[0], score=1.0)
                else:
                    raise ValueError(f"Got list of {len(result)} labels for single-label task.")
            else:
                label_scores: list[tuple[str, float]] = []
                active_set = set(result)
                for label in self._labels:
                    s = 1.0 if label in active_set else 0.0
                    label_scores.append((label, s))
                result = ResultMultiLabel(label_scores=label_scores)

        assert isinstance(result, ResultSingleLabel | ResultMultiLabel)

        scores = self._result_to_scores(result)
        if self._mode == "multi":
            return {"score_per_label": scores}
        else:
            # For single label, it returns a dict with one key.
            label = list(scores.keys())[0]
            score = scores[label]
            return {"label": label, "score": score}

    @override
    def _evaluate_dspy_example(self, truth: dspy.Example, pred: dspy.Prediction, trace: Any, model: dspy.LM) -> float:
        if self._mode == "single":
            return 1.0 if truth["label"] == pred["label"] else 0.0

        # For multi-label: compute label-wise accuracy as
        # 1 - abs(true score for label - predicted score for label)
        # and normalize the sum of label-wise accuracies over all labels.
        accuracy = 0
        for label, score in truth["score_per_label"].items():
            if label in pred["score_per_label"]:
                pred_score = max(min(pred["score_per_label"][label], 1), 0)
                accuracy += 1 - abs(score - pred_score)

        return accuracy / len(truth["score_per_label"])
